# -*- coding: utf-8 -*-

'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import wx, StringIO # We cannot use CStringIO since unicode strings are used below.
import test
from taskcoachlib import persistence, config, meta
from taskcoachlib.domain import base, task, effort, date, category, note, attachment
from taskcoachlib.syncml.config import SyncMLConfigNode


class XMLWriterTest(test.TestCase):
    def setUp(self):
        task.Task.settings = config.Settings(load=False)
        self.fd = StringIO.StringIO()
        self.fd.name = 'testfile.tsk'
        self.fd.encoding = 'utf-8'
        self.writer = persistence.XMLWriter(self.fd)
        self.task = task.Task()
        self.taskList = task.TaskList([self.task])
        self.category = category.Category('Category')
        self.categoryContainer = category.CategoryList([self.category])
        self.note = note.Note()
        self.noteContainer = note.NoteContainer([self.note])
        self.changes = dict()
            
    def __writeAndRead(self):
        self.writer.write(self.taskList, self.categoryContainer, 
            self.noteContainer, SyncMLConfigNode('root'), u'GUID')
        return self.fd.getvalue().decode(self.fd.encoding)
    
    def expectInXML(self, xmlFragment):
        xml = self.__writeAndRead()
        self.failUnless(xmlFragment in xml or \
                        xmlFragment in xml.replace('&apos;', "'"), '%s not in %s'%(xmlFragment, xml))
    
    def expectNotInXML(self, xmlFragment):
        xml = self.__writeAndRead()
        self.failIf(xmlFragment in xml, '%s in %s'%(xmlFragment, xml))
    
    # tests
        
    def testVersion(self):
        self.expectInXML('<?taskcoach release="%s"'%meta.data.version)

    def testGUID(self):
        self.expectInXML('<guid>\nGUID\n</guid>')

    def testTaskSubject(self):
        self.task.setSubject('Subject')
        self.expectInXML('subject="Subject"')

    def testTaskMarkedDeleted(self):
        self.task.markDeleted()
        self.expectInXML('status="3"')

    def testTaskSubjectWithUnicode(self):
        self.task.setSubject(u'ï¬Ÿï­Žï­–')
        self.expectInXML(u'subject="ï¬Ÿï­Žï­–"')
            
    def testTaskDescription(self):
        self.task.setDescription('Description')
        self.expectInXML('<description>\nDescription\n</description>\n')
        
    def testEmptyTaskDescriptionIsNotWritten(self):
        self.expectNotInXML('<description>')
        
    def testTaskPlannedStartDateTime(self):
        self.task.setPlannedStartDateTime(date.DateTime(2004,1,1,11,0,0))
        self.expectInXML('plannedstartdate="%s"'%str(self.task.plannedStartDateTime()))
        
    def testNoPlannedStartDateTime(self):
        self.task.setPlannedStartDateTime(date.DateTime())
        self.expectNotInXML('plannedstartdate=')
        
    def testTaskActualStartDateTime(self):
        self.task.setActualStartDateTime(date.DateTime(2007,12,31,9,0,0))
        self.expectInXML('actualstartdate="%s"'%str(self.task.actualStartDateTime()))

    def testNoActualStartDateTime(self):
        self.task.setActualStartDateTime(date.DateTime())
        self.expectNotInXML('actualstartdate=')
         
    def testTaskDueDateTime(self):
        self.task.setDueDateTime(date.DateTime(2004,1,1,10,5,5))
        self.expectInXML('duedate="%s"'%str(self.task.dueDateTime()))

    def testNoDueDateTime(self):
        self.expectNotInXML('duedate=')
                
    def testTaskCompletionDateTime(self):
        self.task.setCompletionDateTime(date.DateTime(2004,1,1,10,8,4))
        self.expectInXML('completiondate="%s"'%str(self.task.completionDateTime()))

    def testNoCompletionDateTime(self):
        self.expectNotInXML('completiondate=')
        
    def testChildTask(self):
        self.task.addChild(task.Task(subject='child'))
        self.expectInXML('subject="child" />\n</task>\n<category')

    def testEffort(self):
        taskEffort = effort.Effort(self.task, date.DateTime(2004,1,1),
            date.DateTime(2004,1,2), description='description\nline 2')
        self.task.addEffort(taskEffort)
        self.expectInXML('<effort id="%s" start="%s" status="%d" stop="%s">\n'
                         '<description>\ndescription\nline 2\n</description>\n'
                         '</effort>'% \
            (taskEffort.id(), taskEffort.getStart(), base.SynchronizedObject.STATUS_NEW, taskEffort.getStop()))
        
    def testThatEffortTimesDoNotContainMilliseconds(self):
        self.task.addEffort(effort.Effort(self.task, 
            date.DateTime(2004,1,1,10,0,0,123456), 
            date.DateTime(2004,1,1,10,0,10,654310)))
        self.expectInXML('start="2004-01-01 10:00:00"')
        self.expectInXML('stop="2004-01-01 10:00:10"')
        
    def testThatEffortStartAndStopAreNotEqual(self):
        self.task.addEffort(effort.Effort(self.task, 
            date.DateTime(2004,1,1,10,0,0,123456), 
            date.DateTime(2004,1,1,10,0,0,654310)))
        self.expectInXML('start="2004-01-01 10:00:00"')
        self.expectInXML('stop="2004-01-01 10:00:01"')
            
    def testEmptyEffortDescriptionIsNotWritten(self):
        self.task.addEffort(effort.Effort(self.task, date.DateTime(2004,1,1),
            date.DateTime(2004,1,2)))
        self.expectNotInXML('<description>')
        
    def testActiveEffort(self):
        self.task.addEffort(effort.Effort(self.task, date.DateTime(2004,1,1)))
        self.expectInXML('<effort id="%s" start="%s" status="%d" />'%(self.task.efforts()[0].id(), self.task.efforts()[0].getStart(), base.SynchronizedObject.STATUS_NEW))
                
    def testNoEffortByDefault(self):
        self.expectNotInXML('<efforts>')
        
    def testBudget(self):
        self.task.setBudget(date.ONE_HOUR)
        self.expectInXML('budget="%s"'%str(self.task.budget()))
        
    def testNoBudget(self):
        self.expectNotInXML('budget')
        
    def testBudget_MoreThan24Hour(self):
        self.task.setBudget(date.TimeDelta(hours=25))
        self.expectInXML('budget="25:00:00"')
        
    def testOneCategoryWithoutTask(self):
        aCategory = category.Category('test', id="id")
        self.categoryContainer.append(aCategory)
        self.expectInXML('<category creationDateTime="%s" id="id" status="1" '
                         'subject="test" />' % aCategory.creationDateTime())
    
    def testOneCategoryWithOneTask(self):
        self.categoryContainer.append(category.Category('test', [self.task]))
        self.expectInXML('categorizables="%s"' % self.task.id())
        
    def testTwoCategoriesWithOneTask(self):
        subjects = ['test', 'another']
        expectedResults = []
        for subject in subjects:
            cat = category.Category(subject, [self.task])
            self.categoryContainer.append(cat)
            expectedResults.append('<category categorizables="%s" '
                                   'creationDateTime="%s" id="%s" status="1" '
                                   'subject="%s" />' % (self.task.id(), 
                                                        cat.creationDateTime(), 
                                                        cat.id(), subject))
        for expectedResult in expectedResults:
            self.expectInXML(expectedResult)
        
    def testOneCategoryWithSubTask(self):
        child = task.Task()
        self.taskList.append(child)
        self.task.addChild(child)
        self.categoryContainer.append(category.Category('test', [child]))
        self.expectInXML('categorizables="%s"' % child.id())
        
    def testSubCategoryWithoutTasks(self):
        parent = category.Category(subject='parent')
        child = category.Category(subject='child')
        parent.addChild(child)
        self.categoryContainer.extend([parent, child])
        self.expectInXML('<category creationDateTime="%s" id="%s" '
                         'status="1" subject="parent">\n'
                         '<category creationDateTime="%s" id="%s" status="1" '
                         'subject="child" />\n</category>'%\
                         (parent.creationDateTime(), parent.id(), 
                          child.creationDateTime(), child.id()))

    def testSubCategoryWithOneTask(self):
        parent = category.Category(subject='parent')
        child = category.Category(subject='child', categorizables=[self.task])
        parent.addChild(child)
        self.categoryContainer.extend([parent, child])
        self.expectInXML('<category creationDateTime="%s" id="%s" '
                         'status="1" subject="parent">\n'
                         '<category categorizables="%s" creationDateTime="%s" '
                         'id="%s" status="1" subject="child" />\n'
                         '</category>' % (parent.creationDateTime(), parent.id(), 
                                          self.task.id(), child.creationDateTime(),
                                          child.id()))
    
    def testFilteredCategory(self):
        self.categoryContainer.extend([category.Category(subject='test', 
                                                         filtered=True)])
        self.expectInXML('filtered="True"')

    def testCategoryWithDescription(self):
        aCategory = category.Category(subject='subject', description='Description', id='id')
        self.categoryContainer.append(aCategory)
        self.expectInXML('<category creationDateTime="%s" id="id" status="1" subject="subject">\n'
                         '<description>\nDescription\n</description>\n'
                         '</category>' % str(aCategory.creationDateTime()))
        
    def testCategoryWithUnicodeSubject(self):
        unicodeCategory = category.Category(subject=u'ï¬Ÿï­Žï­–', id='id')
        self.categoryContainer.extend([unicodeCategory])
        self.expectInXML(u'subject="ï¬Ÿï­Žï­–"')

    def testCategoryWithDeletedTask(self):
        aCategory = category.Category(subject='category', 
                                      categorizables=[self.task], id='id')
        self.categoryContainer.append(aCategory)
        self.taskList.remove(self.task)
        self.expectInXML('<category creationDateTime="%s" id="id" status="1" '
                         'subject="category" />' % str(aCategory.creationDateTime()))
 
    def testDefaultPriority(self):
        self.expectNotInXML('priority')
        
    def testPriority(self):
        self.task.setPriority(5)
        self.expectInXML('priority="5"')
        
    def testTaskId(self):
        self.expectInXML('id="%s"'%self.task.id())
        
    def testCategoryId(self):
        aCategory = category.Category(subject='category')
        self.categoryContainer.append(aCategory)
        self.expectInXML('id="%s"'%aCategory.id())

    def testNoteId(self):
        self.expectInXML('id="%s"'%self.note.id())

    def testTwoTasks(self):
        self.task.setSubject('task 1')
        task2 = task.Task(subject='task 2')
        self.taskList.append(task2)
        self.expectInXML('subject="task 2"')

    def testDefaultHourlyFee(self):
        self.expectNotInXML('hourlyFee')
        
    def testHourlyFee(self):
        self.task.setHourlyFee(100)
        self.expectInXML('hourlyFee="100"')
        
    def testDefaultFixedFee(self):
        self.expectNotInXML('fixedFee')
        
    def testFixedFee(self):
        self.task.setFixedFee(1000)
        self.expectInXML('fixedFee="1000"')

    def testNoReminder(self):
        self.expectNotInXML('reminder')
        
    def testReminder(self):
        self.task.setReminder(date.DateTime(2005, 5, 7, 13, 15, 10))
        self.expectInXML('reminder="%s"'%str(self.task.reminder()))
        self.expectNotInXML('reminderBeforeSnooze')
        
    def testSnoozedReminder(self):
        now = date.Now()
        self.task.setReminder(now + date.TimeDelta(seconds=30))
        self.task.snoozeReminder(date.TimeDelta(seconds=120), now=lambda: now)
        self.expectInXML('reminder="%s"'%str(self.task.reminder()))
        self.expectInXML('reminderBeforeSnooze="%s"'%str(self.task.reminder(includeSnooze=False)))
        
    def testReminderIsNoneButSnoozedReminderNot(self):
        now = date.Now()
        self.task.setReminder(now + date.TimeDelta(seconds=30))
        self.task.snoozeReminder(date.TimeDelta())
        self.expectNotInXML('reminder')
        
    def testMarkCompletedWhenAllChildrenAreCompletedSetting_None(self):
        self.expectNotInXML('shouldMarkCompletedWhenAllChildrenCompleted')
            
    def testMarkCompletedWhenAllChildrenAreCompletedSetting_True(self):
        self.task.setShouldMarkCompletedWhenAllChildrenCompleted(True)
        self.expectInXML('shouldMarkCompletedWhenAllChildrenCompleted="True"')
            
    def testMarkCompletedWhenAllChildrenAreCompletedSetting_False(self):
        self.task.setShouldMarkCompletedWhenAllChildrenCompleted(False)
        self.expectInXML('shouldMarkCompletedWhenAllChildrenCompleted="False"')
              
    def testNote(self):
        aNote = note.Note(id='id')
        self.noteContainer.append(aNote)
        self.expectInXML('<note creationDateTime="%s" id="id" status="%d" '
                         '/>' % (aNote.creationDateTime(), 
                                 base.SynchronizedObject.STATUS_NEW))
        
    def testNoteWithSubject(self):
        self.noteContainer.append(note.Note(subject='Note'))
        self.expectInXML('subject="Note"')
        
    def testNoteWithDescription(self):
        self.noteContainer.append(note.Note(description='Description'))
        self.expectInXML('<description>\nDescription\n</description>\n')
        
    def testNoteWithChild(self):
        child = note.Note(id='child')
        self.note.addChild(child)
        self.noteContainer.append(child)
        self.expectInXML('<note creationDateTime="%s" id="%s" status="%d">\n'
                         '<note creationDateTime="%s" id="child" status="%d" />\n'
                         '</note>' % (self.note.creationDateTime(), 
                                      self.note.id(),
                                      base.SynchronizedObject.STATUS_NEW,
                                      child.creationDateTime(),
                                      base.SynchronizedObject.STATUS_NEW))
        
    def testNoteWithCategory(self):
        cat = category.Category(subject='cat')
        self.categoryContainer.append(cat)
        self.note.addCategory(cat)
        cat.addCategorizable(self.note)
        self.expectInXML('categorizables="%s"' % self.note.id())

    def testCategoryForegroundColor(self):
        self.categoryContainer.append(category.Category(subject='test', fgColor=wx.RED))
        self.expectInXML('fgColor="(255, 0, 0, 255)"')

    def testCategoryBackgroundColor(self):
        self.categoryContainer.append(category.Category(subject='test', bgColor=wx.RED))
        self.expectInXML('bgColor="(255, 0, 0, 255)"')

    def testDontWriteInheritedCategoryForegroundColor(self):
        parent = category.Category(subject='test', fgColor=wx.RED)
        child = category.Category(subject='child', id='id')
        parent.addChild(child)
        self.categoryContainer.append(parent)
        self.expectInXML('<category creationDateTime="%s" id="id" status="1" '
                         'subject="child" />' % child.creationDateTime())

    def testDontWriteInheritedCategoryBackgroundColor(self):
        parent = category.Category(subject='test', bgColor=wx.RED)
        child = category.Category(subject='child', id='id')
        parent.addChild(child)
        self.categoryContainer.append(parent)
        self.expectInXML('<category creationDateTime="%s" id="id" status="1" '
                         'subject="child" />' % child.creationDateTime())

    def testTaskForegroundColor(self):
        self.task.setForegroundColor(wx.RED)
        self.expectInXML('fgColor="(255, 0, 0, 255)"')
        
    def testTaskBackgroundColor(self):
        self.task.setBackgroundColor(wx.RED)
        self.expectInXML('bgColor="(255, 0, 0, 255)"')

    def testDontWriteInheritedTaskForegroundColor(self):
        self.task.setForegroundColor(wx.RED)
        child = task.Task(subject='child', id='id',
                          plannedStartDateTime=date.DateTime())
        self.task.addChild(child)
        self.taskList.append(child)
        self.expectInXML('<task creationDateTime="%s" id="id" status="1" '
                         'subject="child" />' % child.creationDateTime())
        
    def testDontWriteInheritedTaskBackgroundColor(self):
        self.task.setBackgroundColor(wx.RED)
        child = task.Task(subject='child', id='id', 
                          plannedStartDateTime=date.DateTime())
        self.task.addChild(child)
        self.taskList.append(child)
        self.expectInXML('<task creationDateTime="%s" id="id" status="1" '
                         'subject="child" />' % child.creationDateTime())

    def testNoteForegroundColor(self):
        self.note.setForegroundColor(wx.RED)
        self.expectInXML('fgColor="(255, 0, 0, 255)"')

    def testNoteBackgroundColor(self):
        self.note.setBackgroundColor(wx.RED)
        self.expectInXML('bgColor="(255, 0, 0, 255)"')

    def testDontWriteInheritedNoteForegroundColor(self):
        parent = note.Note(fgColor=wx.RED)
        child = note.Note(subject='child', id='id')
        parent.addChild(child)
        self.noteContainer.append(parent)
        self.expectInXML('<note creationDateTime="%s" id="id" status="1" '
                         'subject="child" />' % child.creationDateTime())
        
    def testDontWriteInheritedNoteBackgroundColor(self):
        parent = note.Note(bgColor=wx.RED)
        child = note.Note(subject='child', id='id')
        parent.addChild(child)
        self.noteContainer.append(parent)
        self.expectInXML('<note creationDateTime="%s" id="id" status="1" '
                         'subject="child" />' % child.creationDateTime())
        
    def testNoRecurencce(self):
        self.expectNotInXML('recurrence')
        
    def testDailyRecurrence(self):
        self.task.setRecurrence(date.Recurrence('daily'))
        self.expectInXML('<recurrence unit="daily" />')
        
    def testWeeklyRecurrence(self):
        self.task.setRecurrence(date.Recurrence('weekly'))
        self.expectInXML('<recurrence unit="weekly" />')

    def testMonthlyRecurrence(self):
        self.task.setRecurrence(date.Recurrence('monthly'))
        self.expectInXML('<recurrence unit="monthly" />')
        
    def testMonthlyRecurrenceOnSameWeekday(self):
        self.task.setRecurrence(date.Recurrence('monthly', sameWeekday=True))
        self.expectInXML('<recurrence sameWeekday="True" unit="monthly" />')

    def testYearlyRecurrence(self):
        self.task.setRecurrence(date.Recurrence('yearly'))
        self.expectInXML('<recurrence unit="yearly" />')
        
    def testRecurrenceCount(self):
        self.task.setRecurrence(date.Recurrence('daily', count=5))
        self.expectInXML('count="5"')

    def testMaxRecurrenceCount(self):
        self.task.setRecurrence(date.Recurrence('daily', maximum=5))
        self.expectInXML('max="5"')

    def testRecurrenceStopDateTime(self):
        stop_datetime = date.DateTime(2000,1,1, 10, 9, 8)
        self.task.setRecurrence(date.Recurrence('daily', 
                                stop_datetime=stop_datetime))
        self.expectInXML('stop_datetime="%s"' % str(stop_datetime))
        
    def testRecurrenceFrequency(self):
        self.task.setRecurrence(date.Recurrence('daily', amount=2))
        self.expectInXML('amount="2"')
        
    def testRecurrenceBasedOnCompletion(self):
        self.task.setRecurrence(date.Recurrence('daily', recurBasedOnCompletion=True))
        self.expectInXML('recurBasedOnCompletion="True"')

    def testNoAttachments(self):
        self.expectNotInXML('attachment')
    
    # addAttachment, addNote, etc., are dynamically generated so pylint can't
    # find them. Disable the error message.
    # pylint: disable=E1101
    
    def testTaskWithOneAttachment(self):
        task_attachment = attachment.FileAttachment('whatever.txt', id='foo')
        self.task.addAttachments(task_attachment)
        self.expectInXML('<attachment creationDateTime="%s" id="foo" '
                         'location="whatever.txt" status="1" '
                         'subject="whatever.txt" type="file" '
                         '/>' % task_attachment.creationDateTime())

    def testObjectWithAttachmentWithNote(self):
        att = attachment.FileAttachment('whatever.txt', id='foo')
        self.task.addAttachments(att)
        attachment_note = note.Note(subject='attnote', id='spam')
        att.addNote(attachment_note)
        self.expectInXML('<attachment creationDateTime="%s" id="foo" '
                         'location="whatever.txt" '
                         'status="1" subject="whatever.txt" type="file">\n'
                         '<note' % att.creationDateTime())

    def testNoteWithOneAttachment(self):
        note_attachment = attachment.FileAttachment('whatever.txt', id='foo')
        self.note.addAttachments(note_attachment)
        self.expectInXML('<attachment creationDateTime="%s" id="foo" '
                         'location="whatever.txt" status="1" '
                         'subject="whatever.txt" type="file" '
                         '/>' % note_attachment.creationDateTime())

    def testCategoryWithOneAttachment(self):
        cat = category.Category('cat')
        self.categoryContainer.append(cat)
        category_attachment = attachment.FileAttachment('whatever.txt', id='foo')
        cat.addAttachments(category_attachment)
        self.expectInXML('<attachment creationDateTime="%s" id="foo" '
                         'location="whatever.txt" status="1" '
                         'subject="whatever.txt" type="file" '
                         '/>' % category_attachment.creationDateTime())
        
    def testTaskWithTwoAttachments(self):
        attachments = [attachment.FileAttachment('whatever.txt'),
                       attachment.FileAttachment('/home/frank/attachment.doc')]
        for a in attachments:
            self.task.addAttachments(a)
        for att in attachments:
            self.expectInXML('<attachment creationDateTime="%s" id="%s" '
                             'location="%s" status="1" subject="%s" type="file" '
                             '/>' % (att.creationDateTime(), att.id(), 
                                     att.location(), att.location()))
        
    def testTaskWithNote(self):
        self.task.addNote(self.note)
        self.expectInXML('>\n<note creationDateTime="%s" id="%s" status="1" '
                         '/>\n</task>' % (self.note.creationDateTime(),
                                          self.note.id()))

    def testTaskWithNotes(self):
        anotherNote = note.Note(subject='Another note', id='id')
        self.task.addNote(self.note)
        self.task.addNote(anotherNote)
        self.expectInXML('>\n<note creationDateTime="%s" id="%s" status="1" />\n'
            '<note creationDateTime="%s" id="id" status="1" subject="Another note" '
            '/>\n</task>' % (self.note.creationDateTime(), self.note.id(),
                             anotherNote.creationDateTime()))
        
    def testTaskWithNestedNotes(self):
        subNote = note.Note(subject='Subnote', id='id')
        self.note.addChild(subNote)
        self.task.addNote(self.note)
        self.expectInXML('>\n<note creationDateTime="%s" id="%s" status="1">\n'
            '<note creationDateTime="%s" id="id" status="1" subject="Subnote" '
            '/>\n</note>\n</task>' % (self.note.creationDateTime(), 
                                      self.note.id(), 
                                      subNote.creationDateTime()))

    def testTaskWithNoteWithCategory(self):
        newNote = note.Note()
        self.task.addNote(newNote)
        newNote.addCategory(self.category)
        self.category.addCategorizable(newNote)
        self.expectInXML('categorizables="%s"' % newNote.id())
  
    def testTaskWithNoteWithSubNoteWithCategory(self):
        newNote = note.Note()
        newSubNote = note.Note()
        newNote.addChild(newSubNote)
        self.task.addNote(newNote)
        newSubNote.addCategory(self.category)
        self.category.addCategorizable(newSubNote)
        self.expectInXML('categorizables="%s"' % newSubNote.id())
        
    def testCategoryWithNote(self):
        self.category.addNote(self.note)
        self.expectInXML('>\n<note creationDateTime="%s" id="%s" status="1" '
                         '/>\n</category>' % (self.note.creationDateTime(), 
                                              self.note.id()))

    def testCategoryWithNotes(self):
        anotherNote = note.Note(subject='Another note', id='id')
        self.category.addNote(self.note)
        self.category.addNote(anotherNote)
        self.expectInXML('>\n<note creationDateTime="%s" id="%s" status="1" />\n'
            '<note creationDateTime="%s" id="id" status="1" subject="Another '
            'note" />\n</category>' % (self.note.creationDateTime(), 
                                       self.note.id(),
                                       anotherNote.creationDateTime()))
        
    def testCategoryWithNestedNotes(self):
        subNote = note.Note(subject='Subnote', id='id')
        self.note.addChild(subNote)
        self.category.addNote(self.note)
        self.expectInXML('>\n<note creationDateTime="%s" id="%s" status="1">\n'
            '<note creationDateTime="%s" id="id" status="1" subject="Subnote" '
            '/>\n</note>\n</category>' % (self.note.creationDateTime(), 
                                          self.note.id(), 
                                          subNote.creationDateTime()))

    def testTaskDefaultExpansionState(self):
        # Don't write anything if the task is not expanded: 
        self.expectNotInXML('expandedContexts')

    def testTaskExpansionState(self):
        self.task.expand()
        self.expectInXML('''expandedContexts="('None',)"''')

    def testTaskExpansionState_SpecificContext(self):
        self.task.expand(context='Test')
        self.expectInXML('''expandedContexts="('Test',)"''')

    def testTaskExpansionState_MultipleContexts(self):
        self.task.expand(context='Test')
        self.task.expand(context='Another context')
        self.expectInXML('''expandedContexts="('Another context', 'Test')"''')

    def testCategoryExpansionState(self):
        cat = category.Category('cat')
        self.categoryContainer.append(cat)
        cat.expand()
        self.expectInXML('''expandedContexts="('None',)"''')

    def testNoteExpansionState(self):
        self.note.expand()
        self.expectInXML('''expandedContexts="('None',)"''')
        
    def testPercentageComplete(self):
        self.task.setPercentageComplete(50)
        self.expectInXML('''percentageComplete="50"''')

    def testPercentageComplete_Float(self):
        self.task.setPercentageComplete(50.0)
        self.expectInXML('''percentageComplete="50.0"''')
        
    def testExclusiveSubcategories(self):
        self.category.makeSubcategoriesExclusive()
        self.expectInXML('''exclusiveSubcategories="True"''')

    def testNonExclusiveSubcategoriesByDefault(self):
        self.expectNotInXML('''exclusiveSubcategories''')
        
    def testTaskFont(self):
        self.task.setFont(wx.SWISS_FONT)
        self.expectInXML('font="%s"'%wx.SWISS_FONT.GetNativeFontInfoDesc())

    def testNoTaskFontByDefault(self):
        self.expectNotInXML('font')
        
    def testNoteFont(self):
        self.note.setFont(wx.SWISS_FONT)
        self.expectInXML('font="%s"'%wx.SWISS_FONT.GetNativeFontInfoDesc())

    def testCategoryFont(self):
        self.category.setFont(wx.SWISS_FONT)
        self.expectInXML('font="%s"'%wx.SWISS_FONT.GetNativeFontInfoDesc())
        
    def testAttachmentFont(self):
        att = attachment.FileAttachment('whatever.txt', id='foo', font=wx.SWISS_FONT)
        self.task.addAttachments(att)
        self.expectInXML('font="%s"'%wx.SWISS_FONT.GetNativeFontInfoDesc())

    def testNonAsciiFontName(self):
        class FakeFont(object):
            def GetNativeFontInfoDesc(self):
                return u'微软雅黑'
        font = FakeFont()
        self.task.setFont(font)
        self.expectInXML(u'font="微软雅黑"')
        
    def testTaskIcon(self):
        self.task.setIcon('icon')
        self.expectInXML('icon="icon"')

    def testNoTaskIcon(self):
        self.expectNotInXML('icon')

    def testSelectedTaskIcon(self):
        self.task.setSelectedIcon('icon')
        self.expectInXML('selectedIcon="icon"')

    def testNoSelectedTaskIcon(self):
        self.expectNotInXML('selectedIcon')

    def testNoteIcon(self):
        self.note.setIcon('icon')
        self.expectInXML('icon="icon"')

    def testSelectedNoteIcon(self):
        self.note.setSelectedIcon('icon')
        self.expectInXML('selectedIcon="icon"')
        
    def testCategoryIcon(self):
        self.category.setIcon('icon')
        self.expectInXML('icon="icon"')

    def testSelectedCategoryIcon(self):
        self.category.setSelectedIcon('icon')
        self.expectInXML('selectedIcon="icon"')

    def testAttachmentIcon(self):
        att = attachment.FileAttachment('whatever.txt', id='foo', icon='icon')
        self.task.addAttachments(att)
        self.expectInXML('icon="icon"')

    def testSelectedAttachmentIcon(self):
        att = attachment.FileAttachment('whatever.txt', selectedIcon='icon')
        self.task.addAttachments(att)
        self.expectInXML('selectedIcon="icon"')

    def testPrerequisite(self):
        prerequisite = task.Task(subject='prereq')
        self.taskList.append(prerequisite)
        self.task.addPrerequisites([prerequisite])
        self.expectInXML('prerequisites="%s"'%prerequisite.id())

    def testMultiplePrerequisites(self):
        # Use the same id's for both prerequisites because we don't know in
        # what order they will end up in the XML.
        prerequisites = [task.Task(subject='prereq1', id='id'),
                         task.Task(subject='prereq2', id='id')]
        self.taskList.extend(prerequisites)
        self.task.addPrerequisites(prerequisites)
        self.expectInXML('prerequisites="id id"')
        
    def testEncodingAttribute(self):
        self.expectInXML('encoding="utf-8"')
        
    def testCreationDateTime(self):
        self.expectInXML('creationDateTime="%s"' % str(self.task.creationDateTime()))
        
    def testDoNotWriteUnknownCreationDateTime(self):
        task_with_unknown_creation_datetime = task.Task(creationDateTime=date.DateTime.min)
        self.taskList.append(task_with_unknown_creation_datetime)
        self.expectNotInXML('creationDateTime="0001-01-01 00:00:00"')
        
    def testModificationDateTime(self):
        self.task.setModificationDateTime(date.DateTime(2013, 1, 1, 0, 0, 0))
        self.expectInXML('modificationDateTime="2013-01-01 00:00:00"')

    def testDoNotWriteUnknownModificationDateTime(self):
        task_with_unknown_modification_datetime = \
            task.Task(modificationDateTime=date.DateTime.min)
        self.expectNotInXML('modificationDateTime="0001-01-01 00:00:00"')



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import wx
from taskcoachlib.i18n import _
from wx.lib import sized_controls
from taskcoachlib import meta, widgets


class ExportDialog(sized_controls.SizedDialog):
    ''' Base class for all export dialogs. Use control classes below to add
        features. '''

    title = 'Override in subclass'
    section = 'export'
    
    def __init__(self, *args, **kwargs):
        self.window = args[0]
        self.settings = kwargs.pop('settings')
        super(ExportDialog, self).__init__(title=self.title, *args, **kwargs)
        pane = self.GetContentsPane()
        pane.SetSizerType('vertical')
        self.components = self.createInterior(pane)
        buttonSizer = self.CreateStdDialogButtonSizer(wx.OK | wx.CANCEL)
        self.SetButtonSizer(buttonSizer)
        buttonSizer.GetAffirmativeButton().Bind(wx.EVT_BUTTON, self.onOk)
        self.Fit()
        self.CentreOnParent()

    def createInterior(self, pane):
        raise NotImplementedError
        
    def exportableViewers(self):
        return self.window.viewer
    
    def activeViewer(self):
        return self.window.viewer.activeViewer()
        
    def options(self):
        result = dict()
        for component in self.components:
            result.update(component.options())
        return result
    
    def onOk(self, event):
        event.Skip()
        for component in self.components:
            component.saveSettings()


# Controls for adding behavior to the base export dialog:

ViewerPickedEvent, EVT_VIEWERPICKED = wx.lib.newevent.NewEvent()


class ViewerPicker(sized_controls.SizedPanel):
    ''' Control for adding a viewer chooser widget to the export dialog. '''
    
    def __init__(self, parent, viewers, activeViewer):
        super(ViewerPicker, self).__init__(parent)
        self.SetSizerType('horizontal')
        self.createPicker()
        self.populatePicker(viewers)
        self.selectActiveViewer(viewers, activeViewer)

    def createPicker(self):
        label = wx.StaticText(self, label=_('Export items from:'))
        label.SetSizerProps(valign='center')
        self.viewerComboBox = wx.ComboBox(self, style=wx.CB_READONLY | wx.CB_SORT)  # pylint: disable=W0201
        self.viewerComboBox.Bind(wx.EVT_COMBOBOX, self.onViewerChanged)
        
    def populatePicker(self, viewers):
        self.titleToViewer = dict()  # pylint: disable=W0201
        for viewer in viewers:
            self.viewerComboBox.Append(viewer.title())  # pylint: disable=E1101
            # Would like to user client data in the combobox, but that 
            # doesn't work on all platforms
            self.titleToViewer[viewer.title()] = viewer
            
    def selectActiveViewer(self, viewers, activeViewer):
        selectedViewer = activeViewer if activeViewer in viewers else viewers[0]
        self.viewerComboBox.SetValue(selectedViewer.title())
        
    def selectedViewer(self):
        return self.titleToViewer[self.viewerComboBox.GetValue()]
    
    def options(self):
        return dict(selectedViewer=self.selectedViewer())

    def onViewerChanged(self, event):        
        event.Skip()
        wx.PostEvent(self, ViewerPickedEvent(viewer=self.selectedViewer()))   
        
    def saveSettings(self):
        pass  # No settings to remember
            

class SelectionOnlyCheckBox(wx.CheckBox):  
    ''' Control for adding a widget to the export dialog that lets the
        user choose between exporting all items or just the selected items. '''
        
    def __init__(self, parent, settings, section, setting):
        super(SelectionOnlyCheckBox, self).__init__(parent, 
            label=_('Export only the selected items'))
        self.settings = settings
        self.section = section
        self.setting = setting
        self.initializeCheckBox()
        
    def initializeCheckBox(self):
        selectionOnly = self.settings.getboolean(self.section, self.setting)
        self.SetValue(selectionOnly)

    def options(self):
        return dict(selectionOnly=self.GetValue())

    def saveSettings(self):
        self.settings.set(self.section, self.setting,  # pylint: disable=E1101
                          str(self.GetValue()))    


class ColumnPicker(sized_controls.SizedPanel):
    ''' Control that lets the user select which columns should be used for 
        exporting. '''

    def __init__(self, parent, viewer):
        super(ColumnPicker, self).__init__(parent)        
        self.SetSizerType('horizontal')
        self.SetSizerProps(expand=True, proportion=1)
        self.createColumnPicker()
        self.populateColumnPicker(viewer)
        
    def createColumnPicker(self):
        label = wx.StaticText(self, label=_('Columns to export:'))
        label.SetSizerProps(valign='top')
        self.columnPicker = widgets.CheckListBox(self)  # pylint: disable=W0201
        self.columnPicker.SetSizerProps(expand=True, proportion=1)
        
    def populateColumnPicker(self, viewer):
        self.columnPicker.Clear()
        self.fillColumnPicker(viewer)
                    
    def fillColumnPicker(self, viewer):
        if not viewer.hasHideableColumns():
            return
        visibleColumns = viewer.visibleColumns()
        for column in viewer.selectableColumns():
            if column.header():
                index = self.columnPicker.Append(column.header(), clientData=column)
                self.columnPicker.Check(index, column in visibleColumns)
            
    def selectedColumns(self):
        indices = [index for index in range(self.columnPicker.GetCount()) if self.columnPicker.IsChecked(index)]
        return [self.columnPicker.GetClientData(index) for index in indices]

    def options(self):
        return dict(columns=self.selectedColumns())
    
    def saveSettings(self):
        pass  # No settings to save


class SeparateDateAndTimeColumnsCheckBox(wx.CheckBox):
    ''' Control that lets the user decide whether dates and times should be 
        separated or kept together. '''
        
    def __init__(self, parent, settings, section, setting):
        super(SeparateDateAndTimeColumnsCheckBox, self).__init__(parent,
            label=_('Put task dates and times in separate columns'))
        self.settings = settings
        self.section = section
        self.setting = setting
        self.initializeCheckBox()
        
    def initializeCheckBox(self):
        separateDateAndTimeColumns = self.settings.getboolean(self.section,
            self.setting)
        self.SetValue(separateDateAndTimeColumns)

    def options(self):
        return dict(separateDateAndTimeColumns=self.GetValue())

    def saveSettings(self):
        self.settings.setboolean(self.section, self.setting, self.GetValue())
        

class SeparateCSSCheckBox(sized_controls.SizedPanel):
    ''' Control to let the user write CSS style information to a 
        separate file instead of including it into the HTML file. '''
    
    def __init__(self, parent, settings, section, setting):
        super(SeparateCSSCheckBox, self).__init__(parent)
        self.settings = settings
        self.section = section
        self.setting = setting
        self.createCheckBox()
        self.createHelpInformation(parent)
        
    def createCheckBox(self):
        self.separateCSSCheckBox = wx.CheckBox(self,  # pylint: disable=W0201
            label=_('Write style information to a separate CSS file'))
        separateCSS = self.settings.getboolean(self.section, self.setting)
        self.separateCSSCheckBox.SetValue(separateCSS)
        
    def createHelpInformation(self, parent):
        width = max([child.GetSize()[0] for child in [self.separateCSSCheckBox] + list(parent.GetChildren())])
        info = wx.StaticText(self, 
            label=_('If a CSS file exists for the exported file, %(name)s will not overwrite it. ' \
                    'This allows you to change the style information without losing your changes on the next export.') % meta.metaDict)
        info.Wrap(width)
        
    def options(self):
        return dict(separateCSS=self.separateCSSCheckBox.GetValue())
        
    def saveSettings(self):
        self.settings.set(self.section, self.setting, 
                          str(self.separateCSSCheckBox.GetValue()))
        

# Export dialogs for different file types:

class ExportAsCSVDialog(ExportDialog):
    title = _('Export as CSV')
    
    def createInterior(self, pane):
        viewerPicker = ViewerPicker(pane, self.exportableViewers(), self.activeViewer())
        viewerPicker.Bind(EVT_VIEWERPICKED, self.onViewerChanged)
        # pylint: disable=W0201
        self.columnPicker = ColumnPicker(pane, viewerPicker.selectedViewer())
        selectionOnlyCheckBox = SelectionOnlyCheckBox(pane, self.settings, 
                                                      self.section, 'csv_selectiononly')
        self.separateDateAndTimeColumnsCheckBox = \
            SeparateDateAndTimeColumnsCheckBox(pane, self.settings, self.section,
                                               'csv_separatedateandtimecolumns')
        self.__check(viewerPicker.selectedViewer())
        return viewerPicker, self.columnPicker, selectionOnlyCheckBox, self.separateDateAndTimeColumnsCheckBox
    
    def onViewerChanged(self, event):
        event.Skip()
        self.columnPicker.populateColumnPicker(event.viewer)
        self.__check(event.viewer)

    def __check(self, viewer):
        self.separateDateAndTimeColumnsCheckBox.Enable(viewer.isShowingTasks() or viewer.isShowingEffort())

          
class ExportAsICalendarDialog(ExportDialog):
    title = _('Export as iCalendar')
    
    def createInterior(self, pane):
        viewerPicker = ViewerPicker(pane, self.exportableViewers(), self.activeViewer())
        selectionOnlyCheckBox = SelectionOnlyCheckBox(pane, self.settings, 
                                                      self.section, 'ical_selectiononly')
        return viewerPicker, selectionOnlyCheckBox

    def exportableViewers(self):
        viewers = super(ExportAsICalendarDialog, self).exportableViewers()
        return [viewer for viewer in viewers if
                viewer.isShowingTasks() or 
                (viewer.isShowingEffort() and not viewer.isShowingAggregatedEffort())]


class ExportAsHTMLDialog(ExportDialog):
    title = _('Export as HTML')
    
    def createInterior(self, pane):
        viewerPicker = ViewerPicker(pane, self.exportableViewers(), self.activeViewer())
        viewerPicker.Bind(EVT_VIEWERPICKED, self.onViewerChanged)
        self.columnPicker = ColumnPicker(pane, viewerPicker.selectedViewer())  # pylint: disable=W0201
        selectionOnlyCheckBox = SelectionOnlyCheckBox(pane, self.settings, self.section, 'html_selectiononly')
        separateCSSChooser = SeparateCSSCheckBox(pane, self.settings, self.section, 'html_separatecss')
        return viewerPicker, self.columnPicker, selectionOnlyCheckBox, separateCSSChooser

    def onViewerChanged(self, event):
        event.Skip()
        self.columnPicker.populateColumnPicker(event.viewer)


class ExportAsTodoTxtDialog(ExportDialog):
    title = _('Export as Todo.txt')
    
    def createInterior(self, pane):
        viewerPicker = ViewerPicker(pane, self.exportableViewers(), self.activeViewer())
        selectionOnlyCheckBox = SelectionOnlyCheckBox(pane, self.settings, 
                                                      self.section, 'todotxt_selectiononly')
        return viewerPicker, selectionOnlyCheckBox
           
    def exportableViewers(self):
        viewers = super(ExportAsTodoTxtDialog, self).exportableViewers()
        return [viewer for viewer in viewers if viewer.isShowingTasks()]



# -*- coding: utf-8 -*-

'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import wx
from taskcoachlib.gui.dialog.preferences import SettingsPageBase
from taskcoachlib import widgets, operating_system
from taskcoachlib.i18n import _


class SyncMLBasePage(SettingsPageBase):
    def __init__(self, iocontroller=None, *args, **kwargs):
        super(SyncMLBasePage, self).__init__(*args, **kwargs)

        self.iocontroller = iocontroller
        self.config = iocontroller.syncMLConfig()

    def get(self, section, name):
        if section == 'access':
            if name in [ 'syncUrl' ]:
                return str(self.config.children()[0]['spds']['syncml']['Conn'].get(name))
            elif name in [ 'username' ]:
                return str(self.config.children()[0]['spds']['syncml']['Auth'].get(name))

        elif section == 'task':
            for child in self.config.children()[0]['spds']['sources'].children():
                if child.name.endswith('Tasks'):
                    return child.get(name)            

        elif section == 'note':
            for child in self.config.children()[0]['spds']['sources'].children():
                if child.name.endswith('Notes'):
                    return child.get(name)
        return ''

    def set(self, section, name, value):
        if section == 'access':
            if name in [ 'syncUrl' ]:
                self.config.children()[0]['spds']['syncml']['Conn'].set(name, value)
            elif name in [ 'username' ]:
                self.config.children()[0]['spds']['syncml']['Auth'].set(name, value)

        elif section == 'task':
            for child in self.config.children()[0]['spds']['sources'].children():
                if child.name.endswith('Tasks'):
                    child.set(name, value)
                    break

        elif section == 'note':
            for child in self.config.children()[0]['spds']['sources'].children():
                if child.name.endswith('Notes'):
                    child.set(name, value)
                    break

    def ok(self):
        super(SyncMLBasePage, self).ok()

        self.iocontroller.setSyncMLConfig(self.config)


class SyncMLAccessPage(SyncMLBasePage):
    def __init__(self, *args, **kwargs):
        super(SyncMLAccessPage, self).__init__(*args, **kwargs)

        choice = self.addChoiceSetting(None, 'preset', _('SyncML server'), '',
                                       [('0', _('Custom')),
                                        ('1', _('MemoToo (http://www.memotoo.com/)')),
                                        ('2', _('Horde-based'))])[0]
        wx.EVT_CHOICE(choice, wx.ID_ANY, self.OnPresetChanged)

        self.addTextSetting('access', 'syncUrl', _('SyncML server URL'))
        self.addTextSetting('access', 'username', _('User name/ID'))

        checkBox = self.addBooleanSetting('task', 'dosync', _('Enable tasks synchronization'))
        wx.EVT_CHECKBOX(checkBox, wx.ID_ANY, self.OnSyncTaskChanged)
        self.addTextSetting('task', 'uri', _('Tasks database name'))
        self.addChoiceSetting('task', 'preferredsyncmode', _('Preferred synchronization mode'), '',
                              [('TWO_WAY', _('Two way')),
                               ('SLOW', _('Slow')),
                               ('ONE_WAY_FROM_CLIENT', _('One way from client')),
                               ('REFRESH_FROM_CLIENT', _('Refresh from client')),
                               ('ONE_WAY_FROM_SERVER', _('One way from server')),
                               ('REFRESH_FROM_SERVER', _('Refresh from server'))])
        self.enableTextSetting('task', 'uri', self.getboolean('task', 'dosync'))
        self.enableChoiceSetting('task', 'preferredsyncmode', self.getboolean('task', 'dosync'))

        checkBox = self.addBooleanSetting('note', 'dosync', _('Enable notes synchronization'))
        wx.EVT_CHECKBOX(checkBox, wx.ID_ANY, self.OnSyncNoteChanged)
        self.addTextSetting('note', 'uri', _('Notes database name'))
        self.addChoiceSetting('note', 'preferredsyncmode', _('Preferred synchronization mode'), '',
                              [('TWO_WAY', _('Two way')),
                               ('SLOW', _('Slow')),
                               ('ONE_WAY_FROM_CLIENT', _('One way from client')),
                               ('REFRESH_FROM_CLIENT', _('Refresh from client')),
                               ('ONE_WAY_FROM_SERVER', _('One way from server')),
                               ('REFRESH_FROM_SERVER', _('Refresh from server'))])
        self.enableTextSetting('note', 'uri', self.getboolean('note', 'dosync'))
        self.enableChoiceSetting('note', 'preferredsyncmode', self.getboolean('note', 'dosync'))

        self.fit()

    def OnPresetChanged(self, event):
        if event.GetInt() == 1:
            self.setTextSetting('access', 'syncUrl', 'http://sync.memotoo.com/syncml')
            self.setTextSetting('task', 'uri', 'task')
            self.setTextSetting('note', 'uri', 'note')
        elif event.GetInt() == 2:
            self.setTextSetting('access', 'syncUrl', 'http://<hostname_and_path>/rpc.php')
            self.setTextSetting('task', 'uri', 'tasks')
            self.setTextSetting('note', 'uri', 'notes')

    def OnSyncTaskChanged(self, event):
        self.enableTextSetting('task', 'uri', event.IsChecked())
        self.enableChoiceSetting('task', 'preferredsyncmode', event.IsChecked())

    def OnSyncNoteChanged(self, event):
        self.enableTextSetting('note', 'uri', event.IsChecked())
        self.enableChoiceSetting('note', 'preferredsyncmode', event.IsChecked())


class SyncMLPreferences(widgets.NotebookDialog):
    def __init__(self, iocontroller=None, *args, **kwargs):
        self.iocontroller = iocontroller
        super(SyncMLPreferences, self).__init__(bitmap='wrench_icon', *args,
                                                **kwargs)
        self.SetSize((700, -1))
        if operating_system.isMac():
            self.CentreOnParent()

    def addPages(self):
        self._interior.SetMinSize((550, 400))
        kwargs = dict(parent=self._interior, columns=3,
                      iocontroller=self.iocontroller)
        # pylint: disable=W0142
        pages = [(SyncMLAccessPage(growableColumn=1, **kwargs), _('Access'), 'earth_blue_icon')]

        for page, title, bitmap in pages:
            self._interior.AddPage(page, title, bitmap=bitmap)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import wx
from taskcoachlib.i18n import _


class DirectoryChooser(wx.Panel):
    def __init__(self, *args, **kwargs):
        super(DirectoryChooser, self).__init__(*args, **kwargs)

        self.chooser = wx.DirPickerCtrl(self, wx.ID_ANY, u'')
        self.checkbx = wx.CheckBox(self, wx.ID_ANY, _('None'))

        sz = wx.BoxSizer(wx.VERTICAL)
        sz.Add(self.chooser, 1, wx.EXPAND)
        sz.Add(self.checkbx, 1)

        self.SetSizer(sz)
        self.Fit()

        wx.EVT_CHECKBOX(self.checkbx, wx.ID_ANY, self.OnCheck)

    def SetPath(self, pth):
        if pth:
            self.checkbx.SetValue(False)
            self.chooser.Enable(True)
            self.chooser.SetPath(pth)
        else:
            self.checkbx.SetValue(True)
            self.chooser.SetPath(u'')
            self.chooser.Enable(False)

    def GetPath(self):
        if not self.checkbx.GetValue():
            return self.chooser.GetPath()
        return u''

    def OnCheck(self, evt):
        self.chooser.Enable(not evt.IsChecked())
        self.chooser.SetPath('/') # Workaround for a wx bug



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import patterns
from taskcoachlib.thirdparty.pubsub import pub
from taskcoachlib.i18n import _
import wx


class AttributeSync(object):
    ''' Class used for keeping an attribute of a domain object synchronized with
        a control in a dialog. If the user edits the value using the control, 
        the domain object is changed, using the appropriate command. If the 
        attribute of the domain object is changed (e.g. in another dialog) the 
        value of the control is updated. '''
        
    def __init__(self, attributeGetterName, entry, currentValue, items, 
                 commandClass, editedEventType, changedEventType, callback=None, 
                 **kwargs):
        self._getter = attributeGetterName
        self._entry = entry
        self._currentValue = currentValue
        self._items = items
        self._commandClass = commandClass
        self.__commandKwArgs = kwargs
        self.__changedEventType = changedEventType
        self.__callback = callback
        entry.Bind(editedEventType, self.onAttributeEdited)
        if len(items) == 1:
            self.__start_observing_attribute(changedEventType, items[0])
        
    def onAttributeEdited(self, event):
        event.Skip()
        new_value = self.getValue()
        if new_value != self._currentValue:
            self._currentValue = new_value
            commandKwArgs = self.commandKwArgs(new_value)
            self._commandClass(None, self._items, **commandKwArgs).do()  # pylint: disable=W0142
            self.__invokeCallback(new_value)

    def onAttributeChanged_Deprecated(self, event):  # pylint: disable=W0613
        if self._entry: 
            new_value = getattr(self._items[0], self._getter)()
            if new_value != self._currentValue:
                self._currentValue = new_value
                self.setValue(new_value)
                self.__invokeCallback(new_value)
        else:
            self.__stop_observing_attribute()
            
    def onAttributeChanged(self, newValue, sender):
        if sender in self._items:
            if self._entry:
                if newValue != self._currentValue:
                    self._currentValue = newValue
                    self.setValue(newValue)
                    self.__invokeCallback(newValue)
            else:
                self.__stop_observing_attribute()
            
    def commandKwArgs(self, new_value):
        self.__commandKwArgs['newValue'] = new_value
        return self.__commandKwArgs
    
    def setValue(self, new_value):
        self._entry.SetValue(new_value)
            
    def getValue(self):
        return self._entry.GetValue()

    def __invokeCallback(self, value):
        if self.__callback is not None:
            try:
                self.__callback(value)
            except Exception, e:
                wx.MessageBox(unicode(e), _('Error'), wx.OK)

    def __start_observing_attribute(self, eventType, eventSource):
        if eventType.startswith('pubsub'):
            pub.subscribe(self.onAttributeChanged, eventType)
        else:
            patterns.Publisher().registerObserver(self.onAttributeChanged_Deprecated,
                                                  eventType=eventType,
                                                  eventSource=eventSource)
    
    def __stop_observing_attribute(self):
        try:
            pub.unsubscribe(self.onAttributeChanged, self.__changedEventType)
        except pub.TopicNameError:
            pass
        patterns.Publisher().removeObserver(self.onAttributeChanged_Deprecated)


class FontColorSync(AttributeSync):
    def setValue(self, newValue):
        self._entry.SetColor(newValue)

    def getValue(self):
        return self._entry.GetColor()



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test
from taskcoachlib import config
from taskcoachlib.domain import task, effort, date, category


class DummyTaskList(task.TaskList):
    def __init__(self, *args, **kwargs):
        self.treeMode = 'not set'
        super(DummyTaskList, self).__init__(*args, **kwargs)
        
    def setTreeMode(self, treeMode):
        self.treeMode = treeMode


class TaskSorterTest(test.TestCase):
    def setUp(self):
        a = self.a = task.Task('a')
        b = self.b = task.Task('b')
        c = self.c = task.Task('c')
        d = self.d = task.Task('d')
        self.list = task.TaskList([d, b, c, a])
        self.sorter = task.sorter.Sorter(self.list)

    def testInitiallyEmpty(self):
        sorter = task.sorter.Sorter(task.TaskList())
        self.assertEqual(0, len(sorter))

    def testLength(self):
        self.assertEqual(4, len(self.sorter))

    def testGetItem(self):
        self.assertEqual(self.a, self.sorter[0])

    def testOrder(self):
        self.assertEqual([self.a, self.b, self.c, self.d], list(self.sorter))

    def testRemoveItem(self):
        self.sorter.remove(self.c)
        self.assertEqual([self.a, self.b, self.d], list(self.sorter))
        self.assertEqual(3, len(self.list))

    def testAppend(self):
        e = task.Task('e')
        self.list.append(e)
        self.assertEqual(5, len(self.sorter))
        self.assertEqual(e, self.sorter[-1])

    def testChange(self):
        self.a.setSubject('z')
        self.assertEqual([self.b, self.c, self.d, self.a], list(self.sorter))


class TaskSorterSettingsTest(test.TestCase):        
    def setUp(self):
        self.taskList = task.TaskList()
        self.sorter = task.sorter.Sorter(self.taskList)    
        self.task1 = task.Task(subject='A')
        self.task2 = task.Task(subject='B')
        self.taskList.extend([self.task1, self.task2])

    def testSortBySubject(self):
        self.sorter.sortBy('subject')
        self.sorter.sortAscending()
        self.assertEqual([self.task1, self.task2], list(self.sorter))

    def testSortBySubjectDescending(self):
        self.sorter.sortBy('subject')
        self.sorter.sortAscending(False)
        self.assertEqual([self.task2, self.task1], list(self.sorter))
        
    def testSortDueDateTime(self):
        self.task2.setDueDateTime(date.Now() + date.ONE_WEEK)
        self.sorter.sortBy('dueDateTime')
        self.assertEqual([self.task2, self.task1], list(self.sorter))
                
    def testSortBySubject_TurnOff(self):
        self.task2.setDueDateTime(date.Now() + date.ONE_WEEK)
        self.sorter.sortBy('subject')
        self.sorter.sortBy('dueDateTime')
        self.assertEqual([self.task2, self.task1], list(self.sorter))
        
    def testSortByCompletionStatus(self):
        self.task1.setCompletionDateTime(date.Now())
        self.assertEqual([self.task2, self.task1], list(self.sorter))
        
    def testSortByInactiveStatus(self):
        self.task2.setActualStartDateTime(date.Now())
        self.assertEqual([self.task2, self.task1], list(self.sorter))
    
    def testSortByPlannedStartDateTime(self):
        self.sorter.sortBy('plannedStartDateTime')
        self.task2.setPlannedStartDateTime(date.Tomorrow())
        self.task1.setPlannedStartDateTime(date.Now() + date.ONE_WEEK)
        self.assertEqual([self.task2, self.task1], list(self.sorter))
        
    def testSortByActualStartDateTime(self):
        self.sorter.sortBy('actualStartDateTime')
        self.task1.setActualStartDateTime(date.Yesterday())
        self.task2.setActualStartDateTime(date.Now() - date.ONE_WEEK)
        self.assertEqual([self.task2, self.task1], list(self.sorter))
        
    def testSortByActualStartDateTimeKeepsSortingWhenChangingActualStartDateTime(self):
        self.sorter.sortBy('actualStartDateTime')
        self.task1.setActualStartDateTime(date.Yesterday())
        self.task2.setActualStartDateTime(date.Now() - date.ONE_WEEK)
        self.task1.setActualStartDateTime(date.Now() - date.ONE_YEAR)
        self.assertEqual([self.task1, self.task2], list(self.sorter))        

    def testSortByDueDateTimeDescending(self):
        self.task1.setDueDateTime(date.Now() + date.ONE_YEAR)
        self.task2.setDueDateTime(date.Now() + date.ONE_WEEK)
        self.sorter.sortBy('dueDateTime')
        self.sorter.sortAscending(False)
        self.assertEqual([self.task1, self.task2], list(self.sorter))
        
    def testByDueDateWithoutFirstSortingByStatus(self):
        self.task1.setDueDateTime(date.Now() + date.ONE_YEAR)
        self.task2.setDueDateTime(date.Now() + date.ONE_WEEK)
        self.sorter.sortBy('dueDateTime')
        self.sorter.sortByTaskStatusFirst(False)
        self.task2.setCompletionDateTime(date.Now())
        self.assertEqual([self.task2, self.task1], list(self.sorter))

    def testSortByReminderAscending(self):
        self.sorter.sortBy('reminder')
        self.sorter.sortAscending(True)
        self.task2.setReminder(date.Now())
        self.assertEqual([self.task2, self.task1], list(self.sorter))
     
    def testSortByReminderDescending(self):
        self.sorter.sortBy('reminder')
        self.sorter.sortAscending(False)
        self.task2.setReminder(date.Now())
        self.assertEqual([self.task1, self.task2], list(self.sorter))

    def testSortBySubjectWithFirstSortingByStatus(self):
        self.sorter.sortByTaskStatusFirst(True)
        self.sorter.sortBy('subject')
        self.task1.setCompletionDateTime(date.Now())
        self.assertEqual([self.task2, self.task1], list(self.sorter))
        
    def testSortBySubjectWithoutFirstSortingByStatus(self):
        self.sorter.sortByTaskStatusFirst(False)
        self.sorter.sortBy('subject')
        self.sorter.sortAscending()
        self.task1.setCompletionDateTime(date.Now())
        self.assertEqual([self.task1, self.task2], list(self.sorter))
                
    def testSortCaseSensitive(self):
        self.sorter.sortByTaskStatusFirst(False)
        self.sorter.sortCaseSensitive(True)
        self.sorter.sortBy('subject')
        self.sorter.sortAscending()
        task3 = task.Task('a')
        self.taskList.append(task3)
        self.assertEqual([self.task1, self.task2, task3], list(self.sorter))

    def testSortCaseInsensitive(self):
        self.sorter.sortByTaskStatusFirst(False)
        self.sorter.sortCaseSensitive(False)
        self.sorter.sortBy('subject')
        self.sorter.sortAscending()
        task3 = task.Task('a')
        self.taskList.append(task3)
        self.assertEqual([self.task1, task3, self.task2], list(self.sorter))
    
    def testSortByTimeLeftAscending(self):
        self.task1.setDueDateTime(date.Now() + date.ONE_YEAR)
        self.task2.setDueDateTime(date.Now() + date.ONE_WEEK)
        self.sorter.sortAscending(True)
        self.sorter.sortBy('timeLeft')
        self.assertEqual([self.task2, self.task1], list(self.sorter))

    def testSortByTimeLeftDescending(self):
        self.task1.setDueDateTime(date.Now() + date.ONE_YEAR)
        self.task2.setDueDateTime(date.Now() + date.ONE_WEEK)
        self.sorter.sortBy('timeLeft')
        self.sorter.sortAscending(False)
        self.assertEqual([self.task1, self.task2], list(self.sorter))

    def testSortByBudgetAscending(self):
        self.sorter.sortAscending(True)
        self.sorter.sortBy('budget')
        self.task1.setBudget(date.TimeDelta(100))
        self.task2.setBudget(date.TimeDelta(10))
        self.assertEqual([self.task2, self.task1], list(self.sorter))

    def testSortByBudgetDescending(self):
        self.sorter.sortBy('budget')
        self.sorter.sortAscending(False)
        self.task1.setBudget(date.TimeDelta(100))
        self.assertEqual([self.task1, self.task2], list(self.sorter))

    def testSortByTimeSpentAscending(self):
        self.sorter.sortBy('timeSpent')
        self.task2.addEffort(effort.Effort(self.task2, date.DateTime(2005,1,1), 
                                           date.DateTime(2006,1,1)))
        self.task1.addEffort(effort.Effort(self.task1, date.DateTime(2005,1,1), 
                                           date.DateTime(2007,1,1)))
        self.assertEqual([self.task2, self.task1], list(self.sorter))

    def testSortByTimeSpentDescending(self):
        self.sorter.sortAscending(False)
        self.sorter.sortBy('timeSpent')
        self.task1.addEffort(effort.Effort(self.task1,
            date.DateTime(2005,1,1,10,0,0), date.DateTime(2005,1,1,11,0,0)))
        self.assertEqual([self.task1, self.task2], list(self.sorter))

    def testSortByHourlyFeeAscending(self):
        self.sorter.sortAscending(True)
        self.sorter.sortBy('hourlyFee')
        self.task1.setHourlyFee(100)
        self.task2.setHourlyFee(200)
        self.assertEqual([self.task1, self.task2], list(self.sorter))

    def testSortByHourlyFeeDescending(self):
        self.sorter.sortBy('hourlyFee')
        self.sorter.sortAscending(False)
        self.task1.setHourlyFee(100)
        self.task2.setHourlyFee(200)
        self.assertEqual([self.task2, self.task1], list(self.sorter))
        
    def testSortByPrerequisiteAscending(self):
        self.sorter.sortAscending(True)
        self.sorter.sortBy('prerequisites')
        self.task1.addPrerequisites([self.task2])
        self.assertEqual([self.task2, self.task1], list(self.sorter))

    def testSortByPrerequisiteDescending(self):
        self.sorter.sortBy('prerequisites')
        self.sorter.sortAscending(False)
        self.task2.addPrerequisites([self.task1])
        self.assertEqual([self.task2, self.task1], list(self.sorter))
        
    def testSortByRecursivePrerequisiteAscending(self):
        self.sorter.sortAscending(True)
        self.sorter.sortBy('prerequisites')
        child1 = task.Task(subject='Child 1')
        self.task1.addChild(child1)
        self.taskList.append(child1)
        child1.addPrerequisites([self.task2])
        self.task2.addPrerequisites([self.task1])
        self.assertEqual([self.task1, self.task2, child1], list(self.sorter))
        
    def testSortByDependencyAscending(self):
        self.sorter.sortAscending(True)
        self.sorter.sortBy('dependencies')
        self.task1.addDependencies([self.task2])
        self.task2.addDependencies([self.task1])
        self.assertEqual([self.task2, self.task1], list(self.sorter))

    def testSortByDependencyDescending(self):
        self.sorter.sortBy('dependencies')
        self.sorter.sortAscending(False)
        self.task1.addDependencies([self.task2])
        self.task2.addDependencies([self.task1])
        self.assertEqual([self.task1, self.task2], list(self.sorter))

    def testSortByRecursiveDependencyAscending(self):
        self.sorter.sortAscending(True)
        self.sorter.sortBy('dependencies')
        child1 = task.Task(subject='Child 1')
        self.task1.addChild(child1)
        self.taskList.append(child1)
        child1.addDependencies([self.task2])
        self.task2.addDependencies([self.task1])
        self.assertEqual([self.task1, self.task2, child1], list(self.sorter))
        
    def testAlwaysKeepSubscriptionToCompletionDateTime(self):
        ''' TaskSorter should keep a subscription to task.completionDateTime 
            even when the completion date is not the sort key, because sorting
            on task status (active, completed, etc.) depends on the completion
            date. '''
        
        self.sorter.sortBy('subject')
        self.sorter.sortAscending()
        self.assertEqual([self.task1, self.task2], list(self.sorter))
        self.task1.setCompletionDateTime(date.Now())
        self.assertEqual([self.task2, self.task1], list(self.sorter))

    def testAlwaysKeepSubscriptionToPlannedStartDateTime(self):
        ''' TaskSorter should keep a subscription to task.plannedStartDateTime 
            even when the planned start date is not the sort key, because sorting
            on task status (active, completed, etc.) depends on the planned start
            date. '''
        self.sorter.sortBy('subject')
        self.sorter.sortAscending()
        self.assertEqual([self.task1, self.task2], list(self.sorter))
        self.task2.setPlannedStartDateTime(date.Now() - date.ONE_SECOND)
        self.assertEqual([self.task2, self.task1], list(self.sorter))
 
    def testAlwaysKeepSubscriptionToActualStartDateTime(self):
        ''' TaskSorter should keep a subscription to task.actualStartDateTime 
            even when the actual start date is not the sort key, because sorting
            on task status (active, completed, etc.) depends on the actual start
            date. '''
        self.sorter.sortBy('subject')
        self.sorter.sortAscending()
        self.assertEqual([self.task1, self.task2], list(self.sorter))
        self.task2.setActualStartDateTime(date.Now())
        self.assertEqual([self.task2, self.task1], list(self.sorter))
        
    def testSortByCategories(self):
        self.sorter.sortBy('categories')
        self.task1.addCategory(category.Category('Category 2'))
        self.task2.addCategory(category.Category('Category 1'))
        self.assertEqual([self.task2, self.task1], list(self.sorter))
        
    def testSortByInvalidSortKey(self):
        self.sorter.sortBy('invalidKey')
        self.assertEqual([self.task1, self.task2], list(self.sorter))
        

class TaskSorterTreeModeTest(test.TestCase):
    def setUp(self):
        task.Task.settings = config.Settings(load=False)
        self.taskList = DummyTaskList()
        self.sorter = task.sorter.Sorter(self.taskList, treeMode=True)        
        self.parent1 = task.Task(subject='parent 1')
        self.child1 = task.Task(subject='child 1')
        self.parent1.addChild(self.child1)
        self.parent2 = task.Task(subject='parent 2')
        self.child2 = task.Task(subject='child 2')
        self.parent2.addChild(self.child2)
        self.taskList.extend([self.parent1, self.parent2])

    def testDefaultSortOrder(self):
        self.assertEqual([self.parent1, self.child1, self.parent2, self.child2],
            list(self.sorter))
        
    def testSortByDueDateTime(self):
        self.sorter.sortBy('dueDateTime')
        self.child2.setDueDateTime(date.Now().endOfDay())
        self.failUnless(list(self.sorter).index(self.parent2) < \
            list(self.sorter).index(self.parent1))

    def testSortByPriority(self):
        self.sorter.sortBy('priority')
        self.sorter.sortAscending(False)
        self.parent1.setPriority(5)
        self.child2.setPriority(10)
        self.failUnless(list(self.sorter).index(self.parent2) < \
            list(self.sorter).index(self.parent1))

    def testSortByCategories_WhenParentsHaveNoCategories(self):
        self.child1.addCategory(category.Category('Category 2'))
        self.child2.addCategory(category.Category('Category 1'))
        self.sorter.sortBy('categories')
        self.failUnless(list(self.sorter).index(self.parent2) < \
            list(self.sorter).index(self.parent1))

    def testSortByCategories_WhenParentCategoryEqualsChildCategoryOfAnotherParent(self):
        category1 = category.Category('Category 1')
        category2 = category.Category('Category 2')
        category3 = category.Category('Category 3')
        self.child1.addCategory(category1)
        self.parent1.addCategory(category3)
        self.parent2.addCategory(category2)
        self.sorter.sortBy('categories')
        self.failUnless(list(self.sorter).index(self.parent2) < \
            list(self.sorter).index(self.parent1))
        
    def testSetSorterToListMode(self):
        self.sorter.setTreeMode(False)
        self.assertEqual([self.child1, self.child2, self.parent1, self.parent2],
            list(self.sorter))

    def testTreeModeDelegation_True(self):
        self.sorter.setTreeMode(True)
        self.assertEqual(True, self.taskList.treeMode)
        
    def testTreeModeDelegation_False(self):
        self.sorter.setTreeMode(False)
        self.assertEqual(False, self.taskList.treeMode)

    def testSortByInvalidSortKey(self):
        self.sorter.sortBy('invalidKey')
        self.assertEqual([self.parent1, self.child1, self.parent2, self.child2], 
                         list(self.sorter))
        
            
class EffortSorterTest(test.TestCase):
    def setUp(self):
        task.Task.settings = config.Settings(load=False)
        self.taskList = task.TaskList()
        self.effortList = effort.EffortList(self.taskList)
        self.sorter = effort.EffortSorter(self.effortList)
        self.task = task.Task('Task')
        self.oldestEffort = effort.Effort(self.task,
            date.DateTime(2004,1,1), date.DateTime(2004,1,2))
        self.newestEffort = effort.Effort(self.task,
            date.DateTime(2004,2,1), date.DateTime(2004,2,2))
        self.task.addEffort(self.oldestEffort)
        self.task.addEffort(self.newestEffort)
        self.taskList.append(self.task)

    def testDescending(self):
        self.assertEqual([self.newestEffort, self.oldestEffort], self.sorter)

    def testResort(self):
        self.oldestEffort.setStart(date.DateTime(2004,3,1))
        self.assertEqual([self.oldestEffort, self.newestEffort], self.sorter)

    def testCreateWhenEffortListIsFilled(self):
        sorter = effort.EffortSorter(self.effortList)
        self.assertEqual([self.newestEffort, self.oldestEffort], sorter)

    def testAddEffort(self):
        evenNewerEffort = effort.Effort(self.task,
            date.DateTime(2005,1,1), date.DateTime(2005,1,2))
        self.task.addEffort(evenNewerEffort)
        self.assertEqual([evenNewerEffort, self.newestEffort, 
            self.oldestEffort], self.sorter)
        
    def testTaskEffortComesBeforeChildEffort(self):
        child = task.Task('Child')
        child.setParent(self.task)
        self.task.addChild(child)
        self.taskList.append(child)
        childEffort = effort.Effort(child, date.DateTime(2004,1,1), 
                                      date.DateTime(2008,1,2))
        child.addEffort(childEffort)
        self.assertEqual([self.newestEffort, childEffort, self.oldestEffort],
            self.sorter)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Simple test to send each priority level
import logging
logging.basicConfig(level=logging.WARNING)

import os
import unittest
from gntp.test import GNTPTestCase
import gntp
import gntp.config
import gntp.notifier

ICON_URL = "https://www.google.com/intl/en_com/images/srpr/logo3w.png"
ICON_FILE = os.path.join(os.path.dirname(__file__), "growl-icon.png")
CALLBACK_URL = "http://github.com"


class BasicTests(GNTPTestCase):
	def test_mini(self):
		gntp.notifier.mini('Testing gntp.notifier.mini',
			applicationName=self.application
			)

	def test_config(self):
		gntp.config.mini('Testing gntp.config.mini',
			applicationName=self.application
			)

	def test_priority(self):
		for priority in [2, 1, 0, -1, -2]:
			self.assertIsTrue(self._notify(
				description='Priority %s' % priority,
				priority=priority
				))

	def test_english(self):
		self.assertIsTrue(self._notify(description='Hello World'))

	def test_extra(self):
		self.assertIsTrue(self._notify(description='allô'))

	def test_japanese(self):
		self.assertIsTrue(self._notify(description='おはおう'))

	def test_sticky(self):
		self.assertIsTrue(self._notify(sticky=True, description='Sticky Test'))

	def test_unknown_note(self):
		self.assertRaises(AssertionError, self._notify, noteType='Unknown')

	def test_parse_error(self):
		self.assertRaises(gntp.ParseError, gntp.parse_gntp, 'Invalid GNTP Packet')

	def test_url_icon(self):
		self.assertIsTrue(self._notify(
			icon=ICON_URL,
			description='test_url_icon',
			))

	def test_file_icon(self):
		self.assertIsTrue(self._notify(
			icon=open(ICON_FILE, 'rb').read(),
			description='test_file_icon',
			))

	def test_callback(self):
		self.assertIsTrue(self._notify(
			callback=CALLBACK_URL,
			description='Testing Callback',
			))

	#def test_subscribe(self):
	#	self.assertTrue(self.growl.subscribe(
	#		id='unittest-id',
	#		name='example.com',
	#		port=5000,
	#		))

if __name__ == '__main__':
	unittest.main()



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import wx
import taskcoachlib.thirdparty.aui as aui
from taskcoachlib import operating_system


class AuiManagedFrameWithDynamicCenterPane(wx.Frame):
    def __init__(self, *args, **kwargs):
        super(AuiManagedFrameWithDynamicCenterPane, self).__init__(*args, 
                                                                   **kwargs)
        agwStyle = aui.AUI_MGR_DEFAULT | aui.AUI_MGR_ALLOW_ACTIVE_PANE
        if not operating_system.isWindows():
            # With this style on Windows, you can't dock back floating frames
            agwStyle |= aui.AUI_MGR_USE_NATIVE_MINIFRAMES
        self.manager = aui.AuiManager(self, agwStyle)
        self.manager.SetAutoNotebookStyle(aui.AUI_NB_TOP | \
                                          aui.AUI_NB_CLOSE_BUTTON | \
                                          aui.AUI_NB_SUB_NOTEBOOK | \
                                          aui.AUI_NB_SCROLL_BUTTONS)
        self.bindEvents()

    def bindEvents(self):
        for eventType in aui.EVT_AUI_PANE_CLOSE, aui.EVT_AUI_PANE_FLOATING:
            self.manager.Bind(eventType, self.onPaneClosingOrFloating)

    def onPaneClosingOrFloating(self, event):
        pane = event.GetPane()
        dockedPanes = self.dockedPanes()
        if self.isCenterPane(pane) and len(dockedPanes) == 1:
            event.Veto() 
        else:
            event.Skip() 
            if self.isCenterPane(pane):
                if pane in dockedPanes:                
                    dockedPanes.remove(pane)
                dockedPanes[0].Center()
                
    def addPane(self, window, caption, name, floating=False):
        x, y = window.GetPositionTuple()
        x, y = window.ClientToScreenXY(x, y)
        paneInfo = aui.AuiPaneInfo()
        paneInfo = paneInfo.CloseButton(True).Floatable(True).\
            Name(name).Caption(caption).Right().\
            FloatingSize((300, 200)).BestSize((200, 200)).\
            FloatingPosition((x + 30, y + 30)).\
            CaptionVisible().MaximizeButton().DestroyOnClose()
        if floating:
            paneInfo.Float()
        if not self.dockedPanes():
            paneInfo = paneInfo.Center()
        self.manager.AddPane(window, paneInfo)
        self.manager.Update()
        
    def setPaneTitle(self, window, title):
        self.manager.GetPane(window).Caption(title)
       
    def dockedPanes(self):
        return [pane for pane in self.manager.GetAllPanes() \
                if not pane.IsToolbar() and not pane.IsFloating() \
                and not pane.IsNotebookPage()]
        
    def float(self, window):
        self.manager.GetPane(window).Float()

    @staticmethod
    def isCenterPane(pane):
        return pane.dock_direction_get() == aui.AUI_DOCK_CENTER



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib.domain import base
from taskcoachlib.thirdparty.pubsub import pub
import task


class Sorter(base.TreeSorter):
    DomainObjectClass = task.Task  # What are we sorting
    TaskStatusAttributes = ('prerequisites', 'dueDateTime', 
                            'plannedStartDateTime', 'actualStartDateTime', 
                            'completionDateTime')
    
    def __init__(self, *args, **kwargs):
        self.__treeMode = kwargs.pop('treeMode', False)
        self.__sortByTaskStatusFirst = kwargs.pop('sortByTaskStatusFirst', True)
        super(Sorter, self).__init__(*args, **kwargs)
        for eventType in (task.Task.prerequisitesChangedEventType(),
                          task.Task.dueDateTimeChangedEventType(),
                          task.Task.plannedStartDateTimeChangedEventType(),
                          task.Task.actualStartDateTimeChangedEventType(),
                          task.Task.completionDateTimeChangedEventType()):
            pub.subscribe(self.onAttributeChanged, eventType)
    
    def setTreeMode(self, treeMode=True):
        self.__treeMode = treeMode
        try:
            self.observable().setTreeMode(treeMode)
        except AttributeError:
            pass
        self.reset(forceEvent=True)

    def treeMode(self):
        return self.__treeMode
                
    def sortByTaskStatusFirst(self, sortByTaskStatusFirst):
        self.__sortByTaskStatusFirst = sortByTaskStatusFirst
        # We don't need to invoke self.reset() here since when this property is
        # changed, the sort order also changes which in turn will cause 
        # self.reset() to be called.
                                
    def createSortKeyFunction(self, sortKey):
        statusSortKey = self.__createStatusSortKey()
        regularSortKey = super(Sorter, self).createSortKeyFunction(sortKey)
        return lambda task: statusSortKey(task) + [regularSortKey(task)]

    def __createStatusSortKey(self):
        if self.__sortByTaskStatusFirst:
            if self.isAscending():
                return lambda task: [task.completed(), task.inactive()]
            else:
                return lambda task: [not task.completed(), not task.inactive()]
        else:
            return lambda task: []

    def _registerObserverForAttribute(self, attribute):
        # Sorter is always observing task dates and prerequisites because 
        # sorting by status depends on those attributes. Hence we don't need
        # to subscribe to these attributes when they become the sort key.
        if attribute not in self.TaskStatusAttributes:
            super(Sorter, self)._registerObserverForAttribute(attribute)
            
    def _removeObserverForAttribute(self, attribute):
        # See comment at _registerObserverForAttribute.
        if attribute not in self.TaskStatusAttributes:
            super(Sorter, self)._removeObserverForAttribute(attribute)



'''
Task Coach - Your friendly task manager
Copyright (C) 2011 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib.patterns import Observer, ObservableComposite
from taskcoachlib.domain.categorizable import CategorizableCompositeObject
from taskcoachlib.domain.note import NoteOwner
from taskcoachlib.domain.task import Task
from taskcoachlib.domain.effort import Effort
from taskcoachlib.domain.attachment import AttachmentOwner
from taskcoachlib.thirdparty import guid
from taskcoachlib.thirdparty.pubsub import pub


class ChangeMonitor(Observer):
    """
    This class monitors change to object on a per-attribute basis.
    """

    def __init__(self, id_=None):
        super(ChangeMonitor, self).__init__()

        self.__guid = id_ or guid.generate()
        self.__frozen = False
        self.__collections = []

        self.reset()

    def freeze(self):
        self.__frozen = True

    def thaw(self):
        self.__frozen = False

    def guid(self):
        return self.__guid

    def reset(self):
        self._changes = dict()
        self._classes = set()

    def monitorClass(self, klass):
        if klass not in self._classes:
            for name in klass.monitoredAttributes():
                eventType = getattr(klass, '%sChangedEventType' % name)()    
                if eventType.startswith('pubsub'):
                    pub.subscribe(self.onAttributeChanged, eventType)
                else:
                    self.registerObserver(self.onAttributeChanged_Deprecated, eventType)
                self._classes.add(klass)
            if issubclass(klass, ObservableComposite):
                self.registerObserver(self.onChildAdded, klass.addChildEventType())
                self.registerObserver(self.onChildRemoved, klass.removeChildEventType())
            if issubclass(klass, CategorizableCompositeObject):
                self.registerObserver(self.onCategoryAdded, klass.categoryAddedEventType())
                self.registerObserver(self.onCategoryRemoved, klass.categoryRemovedEventType())
            if issubclass(klass, Task):
                pub.subscribe(self.onEffortAddedOrRemoved, Task.effortsChangedEventType())
                pub.subscribe(self.onPrerequisitesChanged, Task.prerequisitesChangedEventType())
            if issubclass(klass, NoteOwner):
                self.registerObserver(self.onOtherObjectAdded, klass.noteAddedEventType())
                self.registerObserver(self.onOtherObjectRemoved, klass.noteRemovedEventType())
            if issubclass(klass, AttachmentOwner):
                self.registerObserver(self.onOtherObjectAdded, klass.attachmentAddedEventType())
                self.registerObserver(self.onOtherObjectRemoved, klass.attachmentRemovedEventType())
            if issubclass(klass, Effort):
                pub.subscribe(self.onEffortTaskChanged, Effort.taskChangedEventType())

    def unmonitorClass(self, klass):
        if klass in self._classes:
            for name in klass.monitoredAttributes():
                eventType = getattr(klass, '%sChangedEventType' % name)()    
                if eventType.startswith('pubsub'):
                    pub.unsubscribe(self.onAttributeChanged, eventType)
                else:
                    self.removeObserver(self.onAttributeChanged_Deprecated, eventType)
            if issubclass(klass, ObservableComposite):
                self.removeObserver(self.onChildAdded, klass.addChildEventType())
                self.removeObserver(self.onChildRemoved, klass.removeChildEventType())
            if issubclass(klass, CategorizableCompositeObject):
                self.removeObserver(self.onCategoryAdded, klass.categoryAddedEventType())
                self.removeObserver(self.onCategoryRemoved, klass.categoryRemovedEventType())
            if issubclass(klass, Task):
                pub.unsubscribe(self.onEffortAddedOrRemoved, Task.effortsChangedEventType())
                pub.unsubscribe(self.onPrerequisitesChanged, Task.prerequisitesChangedEventType())
            if issubclass(klass, NoteOwner):
                self.removeObserver(self.onOtherObjectAdded, klass.noteAddedEventType())
                self.removeObserver(self.onOtherObjectRemoved, klass.noteRemovedEventType())
            if issubclass(klass, AttachmentOwner):
                self.removeObserver(self.onOtherObjectAdded, klass.attachmentAddedEventType())
                self.removeObserver(self.onOtherObjectRemoved, klass.attachmentRemovedEventType())
            if issubclass(klass, Effort):
                pub.unsubscribe(self.onEffortTaskChanged, Effort.taskChangedEventType())
            self._classes.remove(klass)

    def monitorCollection(self, collection):
        self.__collections.append(collection)
        self.registerObserver(self.onObjectAdded, collection.addItemEventType(), eventSource=collection)
        self.registerObserver(self.onObjectRemoved, collection.removeItemEventType(), eventSource=collection)

    def unmonitorCollection(self, collection):
        self.__collections.remove(collection)
        self.removeObserver(self.onObjectAdded, collection.addItemEventType(), eventSource=collection)
        self.removeObserver(self.onObjectRemoved, collection.removeItemEventType(), eventSource=collection)

    def onAttributeChanged(self, newValue, sender, topic=pub.AUTO_TOPIC):
        if self.__frozen:
            return

        for name in sender.monitoredAttributes():
            if name in topic.getNameTuple():
                if sender.id() in self._changes and self._changes[sender.id()] is not None:
                    self._changes[sender.id()].add(name)

    def onAttributeChanged_Deprecated(self, event):
        if self.__frozen:
            return

        for type_, valBySource in event.sourcesAndValuesByType().items():
            for obj in valBySource.keys():
                for name in obj.monitoredAttributes():
                    if type_ == getattr(obj, '%sChangedEventType' % name)():
                        if obj.id() in self._changes and self._changes[obj.id()] is not None:
                            self._changes[obj.id()].add(name)
                                
    def _objectAdded(self, obj):
        if obj.id() in self._changes:
            if self._changes[obj.id()] is not None and \
                   '__del__' in self._changes[obj.id()]:
                self._changes[obj.id()].remove('__del__')
        else:
            self._changes[obj.id()] = None

    def _objectsAdded(self, event):
        for obj in event.values():
            self._objectAdded(obj)

    def _objectRemoved(self, obj):
        if obj.id() in self._changes:
            if self._changes[obj.id()] is None:
                del self._changes[obj.id()]
            else:
                self._changes[obj.id()].add('__del__')

    def _objectsRemoved(self, event):
        for obj in event.values():
            self._objectRemoved(obj)

    def onChildAdded(self, event):
        if self.__frozen:
            return

        self._objectsAdded(event)
        for obj in event.values():
            if self._changes[obj.id()] is not None:
                self._changes[obj.id()].add('__parent__')

    def onChildRemoved(self, event):
        if self.__frozen:
            return

        self._objectsRemoved(event)
        for obj in event.values():
            if obj in self._changes and self._changes[obj.id()] is not None:
                self._changes[obj.id()].add('__parent__')

    def onObjectAdded(self, event):
        if self.__frozen:
            return

        self._objectsAdded(event)

    def onObjectRemoved(self, event):
        if self.__frozen:
            return

        self._objectsRemoved(event)

    def onOtherObjectAdded(self, event):
        if self.__frozen:
            return

        self._objectsAdded(event)

    def onOtherObjectRemoved(self, event):
        if self.__frozen:
            return

        self._objectsRemoved(event)

    def onEffortAddedOrRemoved(self, newValue, sender):
        efforts, oldValue = newValue
        effortsToAdd = [effort for effort in efforts if effort not in oldValue]
        effortsToRemove = [effort for effort in oldValue if effort not in efforts]
        for effort in effortsToAdd:
            self._objectAdded(effort)
        for effort in effortsToRemove:
            self._objectRemoved(effort)
            
    def onEffortTaskChanged(self, newValue, sender):
        changes = self._changes.get(sender.id(), None)
        if changes is not None:
            changes.add('__task__')

    def onCategoryAdded(self, event):
        if self.__frozen:
            return

        for obj in event.sources():
            if obj.id() in self._changes and self._changes[obj.id()] is not None:
                for theCategory in event.values(source=obj):
                    name = '_category:%s' % theCategory.id()
                    if '__del' + name in self._changes[obj.id()]:
                        self._changes[obj.id()].remove('__del' + name)
                    else:
                        self._changes[obj.id()].add('__add' + name)

    def onCategoryRemoved(self, event):
        if self.__frozen:
            return

        for obj in event.sources():
            if obj.id() in self._changes and self._changes[obj.id()] is not None:
                for theCategory in event.values(source=obj):
                    name = '_category:%s' % theCategory.id()
                    if '__add' + name in self._changes[obj.id()]:
                        self._changes[obj.id()].remove('__add' + name)
                    else:
                        self._changes[obj.id()].add('__del' + name)

    def onPrerequisitesChanged(self, newValue, sender):  # pylint: disable-msg=W0613
        # Need to check whether the sender is actually in one of the collections we monitor
        # Is this really the best way?
        for collection in self.__collections:
            if sender in collection:
                break
        else:
            return
        if sender.id() in self._changes and self._changes[sender.id()] is not None:
            self._changes[sender.id()].add('__prerequisites__')

    def allChanges(self):
        return self._changes

    def getChanges(self, obj):
        return self._changes.get(obj.id(), None)

    def setChanges(self, id_, changes):
        if changes is None:
            del self._changes[id_]
        else:
            self._changes[id_] = changes

    def isRemoved(self, obj):
        return obj.id() in self._changes and \
               self._changes[obj.id()] is not None and \
               '__del__' in self._changes[obj.id()]

    def resetChanges(self, obj):
        self._changes[obj.id()] = set()

    def addChange(self, obj, name):
        changes = self._changes.get(obj.id(), set())
        changes.add(name)
        self._changes[obj.id()] = changes

    def resetAllChanges(self):
        for id_, changes in self._changes.items():
            if changes is not None and '__del__' in changes:
                del self._changes[id_]
            else:
                self._changes[id_] = set()

    def empty(self):
        self._changes = dict()

    def merge(self, monitor):
        for id_, changes in self._changes.items():
            theirChanges = monitor._changes.get(id_, None)
            if theirChanges is not None:
                changes.update(theirChanges)



'''
Task Coach - Your friendly task manager
Copyright (C) 2011 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib.command import NewEffortCommand, EditEffortStopDateTimeCommand
from taskcoachlib.domain import effort, date
from taskcoachlib.i18n import _
from taskcoachlib.notify import NotificationFrameBase, NotificationCenter
from taskcoachlib.patterns import Observer
from taskcoachlib.powermgt import IdleNotifier
from taskcoachlib.thirdparty.pubsub import pub
from taskcoachlib import render
import wx


class WakeFromIdleFrame(NotificationFrameBase):
    def __init__(self, idleTime, effort, displayedEfforts, *args, **kwargs):
        self._idleTime = idleTime
        self._effort = effort
        self._displayed = displayedEfforts
        self._lastActivity = 0
        super(WakeFromIdleFrame, self).__init__(*args, **kwargs)

    def AddInnerContent(self, sizer, panel):
        idleTimeFormatted = render.dateTime(self._idleTime)
        sizer.Add(wx.StaticText(panel, wx.ID_ANY,
            _('No user input since %s. The following task was\nbeing tracked:') % \
                                idleTimeFormatted))
        sizer.Add(wx.StaticText(panel, wx.ID_ANY,
            self._effort.task().subject()))

        btnNothing = wx.Button(panel, wx.ID_ANY, _('Do nothing'))
        btnStopAt = wx.Button(panel, wx.ID_ANY, _('Stop it at %s') % idleTimeFormatted)
        btnStopResume = wx.Button(panel, wx.ID_ANY, _('Stop it at %s and resume now') % idleTimeFormatted)

        sizer.Add(btnNothing, 0, wx.EXPAND | wx.ALL, 1)
        sizer.Add(btnStopAt, 0, wx.EXPAND | wx.ALL, 1)
        sizer.Add(btnStopResume, 0, wx.EXPAND | wx.ALL, 1)

        wx.EVT_BUTTON(btnNothing, wx.ID_ANY, self.DoNothing)
        wx.EVT_BUTTON(btnStopAt, wx.ID_ANY, self.DoStopAt)
        wx.EVT_BUTTON(btnStopResume, wx.ID_ANY, self.DoStopResume)

    def CloseButton(self, panel):
        return None

    def DoNothing(self, event):
        self._displayed.remove(self._effort)
        self.DoClose()

    def DoStopAt(self, event):
        self._displayed.remove(self._effort)
        EditEffortStopDateTimeCommand(newValue=self._idleTime, items=[self._effort]).do()
        self.DoClose()

    def DoStopResume(self, event):
        self._displayed.remove(self._effort)
        EditEffortStopDateTimeCommand(newValue=self._idleTime, items=[self._effort]).do()
        NewEffortCommand(items=[self._effort.task()]).do()
        self.DoClose()


class IdleController(Observer, IdleNotifier):
    def __init__(self, mainWindow, settings, effortList):
        self._mainWindow = mainWindow
        self._settings = settings
        self._effortList = effortList
        self._displayed = set()

        super(IdleController, self).__init__()

        self.__tracker = effort.EffortListTracker(self._effortList)
        self.__tracker.subscribe(self.__onTrackedChanged, 'effortlisttracker')

        pub.subscribe(self.poweroff, 'powermgt.off')
        pub.subscribe(self.poweron, 'powermgt.on')

    def __onTrackedChanged(self, efforts):
        if len(efforts):
            self.resume()
        else:
            self.pause()

    def getMinIdleTime(self):
        return self._settings.getint('feature', 'minidletime') * 60

    def wake(self, timestamp):
        self._lastActivity = timestamp
        self.OnWake()

    def OnWake(self):
        for effort in self.__tracker.trackedEfforts():
            if effort not in self._displayed:
                self._displayed.add(effort)
                frm = WakeFromIdleFrame(date.DateTime.fromtimestamp(self._lastActivity), effort, self._displayed, _('Notification'),
                                        icon=wx.ArtProvider.GetBitmap('taskcoach', wx.ART_FRAME_ICON, (16, 16)))
                NotificationCenter().NotifyFrame(frm)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import wx
from notifier import AbstractNotifier
from taskcoachlib import operating_system


#==============================================================================
# Utils

class AnimatedShow(wx.Timer):
    """
    Utility class to show a frame with an animation
    """

    def __init__(self, frame, show=True):
        super(AnimatedShow, self).__init__()

        if frame.CanSetTransparent():
            self.__frame = frame
            self.__step = 0
            self.__show = show

            id_ = wx.NewId()
            self.SetOwner(self, id_)
            wx.EVT_TIMER(self, id_, self.__OnTick)
            self.Start(100)
            wx.EVT_CLOSE(frame, self.__OnClose)

            frame.SetTransparent(0)

            if show:
                frame.Show()
        else:
            frame.Show(show)

    def __OnTick(self, event): # pylint: disable=W0613
        self.__step += 1

        if self.__show:
            alpha = int(255.0 * self.__step / 10)
        else:
            alpha = int(255.0 * (10 - (self.__step - 1)) / 10)

        self.__frame.SetTransparent(alpha)

        if self.__step == 10:
            self.Stop()

            if not self.__show:
                self.__frame.Close()

    def __OnClose(self, event):
        self.Stop()
        event.Skip()


class AnimatedMove(wx.Timer):
    """
    Utility class to move a frame with an animation
    """

    def __init__(self, frame, destination):
        super(AnimatedMove, self).__init__()

        self.__frame = frame
        self.__origin = frame.GetPositionTuple()
        self.__destination = destination
        self.__step = 0

        id_ = wx.NewId()
        self.SetOwner(self, id_)
        wx.EVT_TIMER(self, id_, self.__OnTick)
        self.Start(100)
        wx.EVT_CLOSE(frame, self.__OnClose)

    def __OnTick(self, event): # pylint: disable=W0613
        x0, y0 = self.__origin
        x1, y1 = self.__destination
        self.__step += 1

        curX = int(x0 + (x1 - x0) * 1.0 * self.__step / 10)
        curY = int(y0 + (y1 - y0) * 1.0 * self.__step / 10)

        self.__frame.SetPosition(wx.Point(curX, curY))

        if self.__step == 10:
            self.Stop()

    def __OnClose(self, event):
        self.Stop()
        event.Skip()


#==============================================================================
# Notifications

if operating_system.isWindows():
    class _NotifyBase(wx.MiniFrame):
        pass
elif operating_system.isGTK():
    class _NotifyBase(wx.PopupWindow):
        def __init__(self, parent, id_, title, style=0): # pylint: disable=W0613,E1003
            super(_NotifyBase, self).__init__(parent, id_) # No style

        def Close(self): # pylint: disable=W0221,E1003
            # Strange...
            super(_NotifyBase, self).Close()
            self.Destroy()
else:
    class _NotifyBase(wx.Frame):
        """FIXME: steals focus..."""


class NotificationFrameBase(_NotifyBase):
    """
    Base class for a notification frame.

    @ivar title: The notification title
    @type title: unicode
    @ivar icon: An optional icon
    @type icon: NoneType or wx.Bitmap
    """

    def __init__(self, title, icon=None, parent=None):
        self.title = title
        self.icon = icon
        style = self.Style() | (wx.STAY_ON_TOP if parent is None else wx.FRAME_FLOAT_ON_PARENT)
        super(NotificationFrameBase, self).__init__(parent, wx.ID_ANY, u'', style=style)
        self.Populate()

    def Populate(self):
        panel = wx.Panel(self, wx.ID_ANY)

        vsz = wx.BoxSizer(wx.VERTICAL)
        hsz = wx.BoxSizer(wx.HORIZONTAL)

        if self.icon is not None:
            hsz.Add(wx.StaticBitmap(panel, wx.ID_ANY, self.icon), 0,
                    wx.ALL|wx.ALIGN_CENTRE, 2)

        # Seems that font copy-on-write does not work sometimes...
        font = wx.FontFromNativeInfoString(wx.NORMAL_FONT.GetNativeFontInfoDesc())
        font.SetPointSize(8)
        font.SetWeight(wx.FONTWEIGHT_BOLD)

        titleCtrl = wx.StaticText(panel, wx.ID_ANY, self.title)
        titleCtrl.SetFont(font)
        hsz.Add(titleCtrl, 1, wx.ALL|wx.ALIGN_CENTRE, 2)

        btn = self.CloseButton(panel)
        if btn is not None:
            hsz.Add(btn, 0, wx.ALL, 2)
            wx.EVT_BUTTON(btn, wx.ID_ANY, self.DoClose)

        vsz.Add(hsz, 0, wx.ALL|wx.EXPAND, 2)

        self.AddInnerContent(vsz, panel)

        panel.SetSizer(vsz)

        sz = wx.BoxSizer()
        sz.Add(panel, 1, wx.EXPAND)
        self.SetSizer(sz)
        self.Fit()

    def Unpopulate(self):
        self.DestroyChildren()
        self.SetSizer(None)

    def CloseButton(self, panel):
        """
        Override this to return a button instance if you want to
        customize the close button. You may also return None but if
        you do so, please provide another way for the user to dismiss
        the notification.
        """

        return wx.BitmapButton(panel, wx.ID_ANY,
                               wx.ArtProvider.GetBitmap('cross_red_icon',
                                                        wx.ART_FRAME_ICON,
                                                        (16, 16)))

    def AddInnerContent(self, sizer, panel):
        """
        Use this to customize the content of the frame.

        @param sizer: A vertical sizer to which you should add
            your UI elements.
        @param panel: Your parent panel.
        """

    def Style(self):
        """Return the frame's style"""

        style = wx.FRAME_NO_TASKBAR|wx.TAB_TRAVERSAL

        if operating_system.isMac():
            # style |= wx.NO_BORDER|wx.POPUP_WINDOW

            # XXXFIXME: without the POPUP_WINDOW style, the frame
            # steals the focus. But with it, when the frame is created
            # while the user is on another Space that the main
            # window's, it cannot receive user events any more, so
            # cannot be dismissed...

            style |= wx.NO_BORDER

        return style

    def DoClose(self, event=None): # pylint: disable=W0613
        """Use this method instead of Close. Never use Close directly."""

        NotificationCenter().HideFrame(self) # pylint: disable=E1101


class NotificationFrame(NotificationFrameBase):
    """
    A simple (and ugly) notification frame that displays a message.

    @ivar message: Message to display
    @type message: unicode
    """

    def __init__(self, message, *args, **kwargs):
        self.message = message

        super(NotificationFrame, self).__init__(*args, **kwargs)

    def AddInnerContent(self, sizer, panel):
        sizer.Add(wx.StaticText(panel, wx.ID_ANY, self.message), 1, wx.ALL|wx.EXPAND, 5)


class _NotificationCenter(wx.EvtHandler):
    """
    The class that handles notification frames.
    """

    framePool = []

    def __init__(self):
        super(_NotificationCenter, self).__init__()

        self.displayedFrames = []
        self.waitingFrames = []
        self.notificationWidth = 300
        self.notificationMargin = 5

        self.__tmr = wx.Timer()
        id_ = wx.NewId()
        self.__tmr.SetOwner(self, id_)
        wx.EVT_TIMER(self, id_, self.__OnTick)
        self.__tmr.Start(1000)

    def NotifyFrame(self, frm, timeout=None):
        """
        Present a new notification frame.

        @param frm: The frame to show
        @param timeout: Time to display the frame before automatically
            hiding it; in seconds.
        """

        if frm.GetParent():
            dx, dy = frm.GetParent().GetPosition()
            dw, dh = frm.GetParent().GetSizeTuple()
        else:
            dx, dy, dw, dh = self.GetDisplayRect()

        w, h = frm.GetSizeTuple()
        w = w if w > self.notificationWidth else self.notificationWidth

        bottom = dy + dh - self.notificationMargin

        for otherFrame, height, tmo in self.displayedFrames:
            bottom -= height + self.notificationMargin
            if bottom - h < 0:
                self.waitingFrames.append((frm, timeout))
                return

        if frm.GetParent():
            x = min(self.GetDisplayRect()[2] - self.notificationMargin - w,
                    dx + dw + self.notificationMargin)
        else:
            x = dx + dw - w - self.notificationMargin

        frm.SetDimensions(x,
                          bottom - h - self.notificationMargin,
                          w, h)
        self.displayedFrames.append((frm, h, timeout))

        frm.Layout()

        AnimatedShow(frm)

    def CheckWaiting(self):
        waiting = self.waitingFrames
        self.waitingFrames = []
        for frm, tmo in waiting:
            self.NotifyFrame(frm, timeout=tmo)

    def Notify(self, title, msg, icon=None, timeout=None):
        """
        Present a new simple notification frame.

        @param title: Notification title
        @param msg: Notification message
        @param timeout: See L{NotifyFrame}.
        """

        frm = NotificationFrame(msg, title, icon=icon)
        self.NotifyFrame(frm, timeout=timeout)

    def HideFrame(self, frm):
        """
        Hide a notification frame.
        """

        for idx, (frame, height, tmo) in enumerate(self.displayedFrames):
            if frame == frm:
                self.displayedFrames[idx] = (frame, height, 1)
                break

        self.CheckWaiting()

    def HideAll(self):
        """
        Hide all notification frames. Call this when you want to close
        your main frame, or else the wx loop won't exit.
        """

        for idx, (frame, height, tmo) in enumerate(self.displayedFrames):
            self.displayedFrames[idx] = (frame, height, 1)
        for frame, timeout in self.waitingFrames:
            frame.Close()
        self.waitingFrames = []

    def GetDisplayRect(self):
        """
        Returns the geometry of the main application frame's display
        """

        dpyIndex = wx.Display.GetFromWindow(wx.GetApp().GetTopWindow())
        if dpyIndex == wx.NOT_FOUND:
            return wx.ClientDisplayRect()
        return wx.Display(dpyIndex).GetClientArea()

    def __OnTick(self, event): # pylint: disable=W0613
        s = 0
        newList = []

        for frame, height, tmo in self.displayedFrames:
            if frame.GetParent():
                dx, dy = frame.GetParent().GetPosition()
                dw, dh = frame.GetParent().GetSizeTuple()
            else:
                dx, dy, dw, dh = self.GetDisplayRect()

            bottom = dy + dh - self.notificationMargin

            if s == 0:
                if tmo == 1:
                    frame.Close()
                    s = 1
                    continue

                newList.append((frame, height, tmo - 1 if tmo is not None else None))
                bottom -= height + self.notificationMargin
            else:
                if tmo == 1:
                    frame.Close()
                else:
                    newList.append((frame, height, tmo - 1 if tmo is not None else None))
                    x, y = frame.GetPositionTuple()
                    AnimatedMove(frame, (x, bottom - height - self.notificationMargin))
                    bottom -= height + self.notificationMargin

        self.displayedFrames = newList
        self.CheckWaiting()


class NotificationCenter(object):
    _instance = None

    def __new__(self):
        if NotificationCenter._instance is None:
            NotificationCenter._instance = _NotificationCenter()
        return NotificationCenter._instance


class UniversalNotifier(AbstractNotifier):
    def getName(self):
        return 'Task Coach'

    def isAvailable(self):
        return True

    def notify(self, title, summary, bitmap, **kwargs):
        NotificationCenter().Notify(title, summary, icon=bitmap) # pylint: disable=E1101


AbstractNotifier.register(UniversalNotifier())


if __name__ == '__main__':
    class TestNotificationFrame(NotificationFrameBase):
        def AddInnerContent(self, sizer, panel):
            choice = wx.Choice(panel, wx.ID_ANY)
            choice.Append('One')
            choice.Append('Two')
            choice.Append('Three')
            sizer.Add(choice, 0, wx.ALL|wx.EXPAND, 5)

            hsz = wx.BoxSizer(wx.HORIZONTAL)
            hsz.Add(wx.Button(panel, wx.ID_ANY, u'OK'), 1, wx.ALL, 2)
            hsz.Add(wx.Button(panel, wx.ID_ANY, u'Cancel'), 1, wx.ALL, 2)
            sizer.Add(hsz, 0, wx.EXPAND|wx.ALL, 5)

        def CloseButton(self, panel):
            return None


    class TestFrame(wx.Frame):
        def __init__(self):
            super(TestFrame, self).__init__(None, wx.ID_ANY, 'Test frame')
            # pylint: disable=E1101
            NotificationCenter().Notify('Sample title', 'Sample content', timeout=3)
            NotificationCenter().Notify('Other sample', 'Multi-line sample content\nfor example\nDont try this at home', timeout=3,
                                        icon=wx.ArtProvider.GetBitmap('taskcoach', wx.ART_FRAME_ICON, (16, 16)))
            NotificationCenter().Notify('Before last sample', 'Spam!')
            NotificationCenter().NotifyFrame(TestNotificationFrame('Test custom',
                                                                   icon=wx.ArtProvider.GetBitmap('taskcoach', wx.ART_FRAME_ICON, (16, 16))))
            NotificationCenter().Notify('Last sample', 'Foobar!')

            wx.EVT_CLOSE(self, self.OnClose)

        def OnClose(self, evt):
            NotificationCenter().HideAll() # pylint: disable=E1101
            evt.Skip()


    class App(wx.App):
        def OnInit(self):
            from taskcoachlib.gui import artprovider
            artprovider.init()
            TestFrame().Show()
            return True


    App(0).MainLoop()



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import os
import test, mock


class SaveTest(test.TestCase):
    def setUp(self):
        self.filename = 'SaveTest.tsk'
        self.filename2 = 'SaveTest2.tsk'
        self.mockApp = mock.App()
        self.mockApp.addTasks()

    def tearDown(self):
        self.mockApp.iocontroller.save()
        self.mockApp.quitApplication()
        for filename in [self.filename, self.filename2]:
            if os.path.isfile(filename):
                os.remove(filename)
        mock.App.deleteInstance()
        super(SaveTest, self).tearDown()
        
    def assertTasksLoaded(self, nrTasks):
        self.assertEqual(nrTasks, len(self.mockApp.taskFile.tasks()))
        
    def testSave(self):
        self.mockApp.iocontroller.saveas(self.filename)
        self.mockApp.iocontroller.open(self.filename)
        self.assertTasksLoaded(2)

    def testSaveSelection_Child(self):
        self.mockApp.iocontroller.saveas(self.filename)
        self.mockApp.iocontroller.saveselection([self.mockApp.child], self.filename2)
        self.mockApp.iocontroller.close()
        self.mockApp.iocontroller.open(self.filename2)
        self.assertTasksLoaded(1)

    def testSaveSelection_Parent(self):
        self.mockApp.iocontroller.saveas(self.filename)
        self.mockApp.iocontroller.saveselection([self.mockApp.parent], self.filename2)
        self.mockApp.iocontroller.close()
        self.mockApp.iocontroller.open(self.filename2)
        self.assertTasksLoaded(2)
        
    def testSaveAndMerge(self):
        mockApp2 = mock.App()
        mockApp2.addTasks()
        mockApp2.iocontroller.saveas(self.filename2)
        self.mockApp.iocontroller.merge(self.filename2)
        self.assertTasksLoaded(4)
        self.mockApp.iocontroller.saveas(self.filename)
        mockApp2.quitApplication()



# -*- coding: utf-8 -*-

'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import patterns
from taskcoachlib.i18n import _
from taskcoachlib.domain import attachment
import base, noteCommands


class EditAttachmentLocationCommand(base.BaseCommand):
    plural_name = _('Edit location of attachments')
    singular_name = _('Edit attachment "%s" location')

    def __init__(self, *args, **kwargs):
        self.__newLocation = kwargs.pop('newValue')
        super(EditAttachmentLocationCommand, self).__init__(*args, **kwargs)
        self.__oldLocations = [item.location() for item in self.items]
    
    @patterns.eventSource
    def do_command(self, event=None):
        super(EditAttachmentLocationCommand, self).do_command()
        for item in self.items:
            item.setLocation(self.__newLocation)
            
    @patterns.eventSource
    def undo_command(self, event=None):
        super(EditAttachmentLocationCommand, self).undo_command()
        for item, oldLocation in zip(self.items, self.__oldLocations):
            item.setLocation(oldLocation)
            
    def redo_command(self):
        self.do_command()


class AddAttachmentCommand(base.BaseCommand):
    plural_name = _('Add attachment')
    singular_name = _('Add attachment to "%s"')
    
    def __init__(self, *args, **kwargs):
        self.owners = []
        self.__attachments = kwargs.get('attachments', 
            [attachment.FileAttachment('', subject=_('New attachment'))])
        super(AddAttachmentCommand, self).__init__(*args, **kwargs)
        self.owners = self.items
        self.items = self.__attachments
        self.save_modification_datetimes()
        
    def modified_items(self):
        return self.owners
        
    @patterns.eventSource
    def addAttachments(self, event=None):
        kwargs = dict(event=event)
        for owner in self.owners:
            owner.addAttachments(*self.__attachments, **kwargs) # pylint: disable=W0142

    @patterns.eventSource
    def removeAttachments(self, event=None):
        kwargs = dict(event=event)
        for owner in self.owners:
            owner.removeAttachments(*self.__attachments, **kwargs) # pylint: disable=W0142
                         
    def do_command(self):
        super(AddAttachmentCommand, self).do_command()
        self.addAttachments()
        
    def undo_command(self):
        super(AddAttachmentCommand, self).undo_command()
        self.removeAttachments()

    def redo_command(self):
        super(AddAttachmentCommand, self).redo_command()
        self.addAttachments()


class RemoveAttachmentCommand(base.BaseCommand):
    plural_name = _('Remove attachment')
    singular_name = _('Remove attachment to "%s"')
    
    def __init__(self, *args, **kwargs):
        self._attachments = kwargs.pop('attachments')
        super(RemoveAttachmentCommand, self).__init__(*args, **kwargs)

    @patterns.eventSource
    def addAttachments(self, event=None):
        kwargs = dict(event=event)
        for item in self.items:
            item.addAttachments(*self._attachments, **kwargs) # pylint: disable=W0142
        
    @patterns.eventSource
    def removeAttachments(self, event=None):
        kwargs = dict(event=event)
        for item in self.items:
            item.removeAttachments(*self._attachments, **kwargs) # pylint: disable=W0142
                
    def do_command(self):
        super(RemoveAttachmentCommand, self).do_command()
        self.removeAttachments()
        
    def undo_command(self):
        super(RemoveAttachmentCommand, self).undo_command()
        self.addAttachments()

    def redo_command(self):
        super(RemoveAttachmentCommand, self).redo_command()
        self.removeAttachments()


class CutAttachmentCommand(base.CutCommandMixin, RemoveAttachmentCommand):
    def itemsToCut(self):
        return self._attachments

    def sourceOfItemsToCut(self):
        class Wrapper(object):
            def __init__(self, items):
                self.__items = items
            def extend(self, attachments):
                for item in self.__items:
                    item.addAttachments(*attachments)
            def removeItems(self, attachments):
                for item in self.__items:
                    item.removeAttachments(*attachments)
        return Wrapper(self.items)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import meta, config, gui, operating_system
from taskcoachlib.domain import task, effort, date
import test


class TaskFileMock(object):
    def filename(self):
        return 'filename'


class MainWindowMock(object):
    taskFile = TaskFileMock()

    def __init__(self):
        self.__cb = None
    
    def restore(self):
        pass  # pragma: no cover

    def Bind(self, evt, cb):
        self.__cb = cb

    def ProcessIdle(self):
        if self.__cb is not None:
            self.__cb(None)


class TaskBarIconTestCase(test.TestCase):
    def setUp(self):
        self.taskList = task.TaskList()
        self.settings = task.Task.settings = config.Settings(load=False)
        self.window = MainWindowMock()
        self.icon = gui.TaskBarIcon(self.window, self.taskList,
            self.settings)

    def tearDown(self):  # pragma: no cover
        if operating_system.isWindows():
            self.icon.Destroy()
        else:
            self.icon.RemoveIcon()
        super(TaskBarIconTestCase, self).tearDown()
            

class TaskBarIconTest(TaskBarIconTestCase):
    def testIcon_NoTasks(self):
        self.window.ProcessIdle()
        self.failUnless(self.icon.IsIconInstalled())
        
    def testStartTracking(self):
        activeTask = task.Task()
        self.taskList.append(activeTask)
        activeTask.addEffort(effort.Effort(activeTask))
        self.assertEqual('clock_icon', self.icon.bitmap())

    def testStopTracking(self):
        activeTask = task.Task()
        self.taskList.append(activeTask)
        activeEffort = effort.Effort(activeTask)
        activeTask.addEffort(activeEffort)
        activeTask.removeEffort(activeEffort)
        self.assertEqual(self.icon.defaultBitmap(), self.icon.bitmap())
        
        
class TaskBarIconTooltipTestCase(TaskBarIconTestCase):
    def assertTooltip(self, text):
        expectedTooltip = '%s - %s' % (meta.name, TaskFileMock().filename()) 
        if text:
            expectedTooltip += '\n%s' % text
        self.assertEqual(expectedTooltip, self.icon.tooltip())


class TaskBarIconTooltipTest(TaskBarIconTooltipTestCase):
    def testNoTasks(self):
        self.assertTooltip('')
        
    def testOneTaskNoDueDateTime(self):
        self.taskList.append(task.Task())
        self.assertTooltip('')

    def testOneTaskDueSoon(self):
        self.taskList.append(task.Task(dueDateTime=date.Now() + date.ONE_HOUR))
        self.assertTooltip('one task due soon')
        
    def testOneTaskNoLongerDueSoonAfterChangingDueSoonSetting(self):
        self.taskList.append(task.Task(dueDateTime=date.Now() + date.ONE_HOUR))
        self.settings.setint('behavior', 'duesoonhours', 0)
        self.assertTooltip('')
        
    def testTwoTasksDueSoon(self):
        self.taskList.append(task.Task(dueDateTime=date.Now() + date.ONE_HOUR))
        self.taskList.append(task.Task(dueDateTime=date.Now() + date.ONE_HOUR))
        self.assertTooltip('2 tasks due soon')
        
    def testOneTasksOverdue(self):
        self.taskList.append(task.Task(dueDateTime=date.Yesterday()))
        self.assertTooltip('one task overdue')
        
    def testTwoTasksOverdue(self):
        self.taskList.append(task.Task(dueDateTime=date.Yesterday()))
        self.taskList.append(task.Task(dueDateTime=date.Yesterday()))
        self.assertTooltip('2 tasks overdue')
        
    def testOneTaskDueSoonAndOneTaskOverdue(self):
        self.taskList.append(task.Task(dueDateTime=date.Yesterday()))
        self.taskList.append(task.Task(dueDateTime=date.Now() + date.ONE_HOUR))
        self.assertTooltip('one task overdue, one task due soon')
        
    def testRemoveTask(self):
        newTask = task.Task()
        self.taskList.append(newTask)
        self.taskList.remove(newTask)
        self.assertTooltip('')
        
    def testRemoveOverdueTask(self):
        overdueTask = task.Task(dueDateTime=date.Yesterday())
        self.taskList.append(overdueTask)
        self.taskList.remove(overdueTask)
        self.assertTooltip('')


class TaskBarIconTooltipWithTrackedTaskTest(TaskBarIconTooltipTestCase):
    def setUp(self):
        super(TaskBarIconTooltipWithTrackedTaskTest, self).setUp()
        self.task = task.Task(subject='Subject')
        self.taskList.append(self.task)
        self.task.addEffort(effort.Effort(self.task))

    def testStartTracking(self):
        self.assertTooltip('tracking "Subject"')

    def testStopTracking(self):
        self.task.efforts()[0].setStop(date.DateTime(2000, 1, 1, 10, 0, 0))
        self.assertTooltip('')

    def testTrackingTwoTasks(self):
        activeTask = task.Task()
        self.taskList.append(activeTask)
        activeTask.addEffort(effort.Effort(activeTask))
        self.assertTooltip('tracking effort for 2 tasks')

    def testChangingSubjectOfTrackedTask(self):
        self.task.setSubject('New subject')
        self.assertTooltip('tracking "New subject"')

    def testChangingSubjectOfTaskThatIsNotTrackedAnymore(self):
        self.task.efforts()[0].setStop(date.DateTime(2000, 1, 1, 10, 0, 0))
        self.task.setSubject('New subject')
        self.assertTooltip('')



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import patterns


def DomainObjectOwnerMetaclass(name, bases, ns):
    """This metaclass makes a class an owner for some domain
    objects. The __ownedType__ attribute of the class must be a
    string. For each type, the following methods will be added to the
    class (here assuming a type of 'Foo'):

      - __init__, __getstate__, __setstate__, __getcopystate__, __setcopystate__
      - addFoo, removeFoo, addFoos, removeFoos
      - setFoos, foos
      - foosChangedEventType
      - fooAddedEventType
      - fooRemovedEventType
      - modificationEventTypes
      - __notifyObservers"""

    # This  metaclass is  a function  instead  of a  subclass of  type
    # because as we're replacing __init__, we don't want the metaclass
    # to be inherited by children.

    klass = type(name, bases, ns)

    def constructor(instance, *args, **kwargs):
        # NB: we use a simple list here. Maybe we should use a container type.
        setattr(instance,'_%s__%ss' % (name, klass.__ownedType__.lower()),
                kwargs.pop(klass.__ownedType__.lower() + 's', []))
        super(klass, instance).__init__(*args, **kwargs)

    klass.__init__ = constructor

    def changedEventType(class_):
        return '%s.%ss' % (class_, klass.__ownedType__.lower())

    setattr(klass, '%ssChangedEventType' % klass.__ownedType__.lower(), 
            classmethod(changedEventType))

    def addedEventType(class_):
        return '%s.%s.added' % (class_, klass.__ownedType__.lower())

    setattr(klass, '%sAddedEventType' % klass.__ownedType__.lower(),
            classmethod(addedEventType))

    def removedEventType(class_):
        return '%s.%s.removed' % (class_, klass.__ownedType__.lower())

    setattr(klass, '%sRemovedEventType' % klass.__ownedType__.lower(),
            classmethod(removedEventType))

    def modificationEventTypes(class_):
        try:
            eventTypes = super(klass, class_).modificationEventTypes()
        except AttributeError:
            eventTypes = []
        return eventTypes + [changedEventType(class_)]
    
    klass.modificationEventTypes = classmethod(modificationEventTypes)

    def objects(instance, recursive=False):
        ownedObjects = getattr(instance, '_%s__%ss' % (name, klass.__ownedType__.lower()))
        result = [ownedObject for ownedObject in ownedObjects \
                if not ownedObject.isDeleted()]
        if recursive:
            for ownedObject in result[:]:
                result.extend(ownedObject.children(recursive=True))
        return result

    setattr(klass, '%ss' % klass.__ownedType__.lower(), objects)

    @patterns.eventSource
    def setObjects(instance, newObjects, event=None):
        if newObjects == objects(instance):
            return
        setattr(instance, '_%s__%ss' % (name, klass.__ownedType__.lower()), 
                                        newObjects)
        changedEvent(instance, event, *newObjects) # pylint: disable=W0142

    setattr(klass, 'set%ss' % klass.__ownedType__, setObjects)

    def changedEvent(instance, event, *objects):
        event.addSource(instance, *objects, 
                        **dict(type=changedEventType(instance.__class__)))
    
    setattr(klass, '%ssChangedEvent' % klass.__ownedType__.lower(), changedEvent)

    def addedEvent(instance, event, *objects):
        event.addSource(instance, *objects,
                        **dict(type=addedEventType(instance.__class__)))

    setattr(klass, '%sAddedEvent' % klass.__ownedType__.lower(), addedEvent)

    def removedEvent(instance, event, *objects):
        event.addSource(instance, *objects,
                        **dict(type=removedEventType(instance.__class__)))

    setattr(klass, '%sRemovedEvent' % klass.__ownedType__.lower(), removedEvent)

    @patterns.eventSource
    def addObject(instance, ownedObject, event=None):
        getattr(instance, '_%s__%ss' % (name, klass.__ownedType__.lower())).append(ownedObject)
        changedEvent(instance, event, ownedObject)
        addedEvent(instance, event, ownedObject)

    setattr(klass, 'add%s' % klass.__ownedType__, addObject)

    @patterns.eventSource
    def addObjects(instance, *ownedObjects, **kwargs):
        if not ownedObjects:
            return
        getattr(instance, '_%s__%ss' % (name, klass.__ownedType__.lower())).extend(ownedObjects)
        event = kwargs.pop('event', None)
        changedEvent(instance, event, *ownedObjects)
        addedEvent(instance, event, *ownedObjects)

    setattr(klass, 'add%ss' % klass.__ownedType__, addObjects)

    @patterns.eventSource
    def removeObject(instance, ownedObject, event=None):
        getattr(instance, '_%s__%ss' % (name, klass.__ownedType__.lower())).remove(ownedObject)
        changedEvent(instance, event, ownedObject)
        removedEvent(instance, event, ownedObject)

    setattr(klass, 'remove%s' % klass.__ownedType__, removeObject)

    @patterns.eventSource
    def removeObjects(instance, *ownedObjects, **kwargs):
        if not ownedObjects:
            return
        for ownedObject in ownedObjects:
            try:
                getattr(instance, '_%s__%ss' % (name, klass.__ownedType__.lower())).remove(ownedObject)
            except ValueError:
                pass
        event = kwargs.pop('event', None)
        changedEvent(instance, event, *ownedObjects)
        removedEvent(instance, event, *ownedObjects)
        
    setattr(klass, 'remove%ss' % klass.__ownedType__, removeObjects)

    def getstate(instance):
        try:
            state = super(klass, instance).__getstate__()
        except AttributeError:
            state = dict()
        state[klass.__ownedType__.lower() + 's'] = getattr(instance, '_%s__%ss' % (name, klass.__ownedType__.lower()))[:]
        return state

    klass.__getstate__ = getstate

    @patterns.eventSource
    def setstate(instance, state, event=None):
        try:
            super(klass, instance).__setstate__(state, event=event)
        except AttributeError:
            pass
        setObjects(instance, state[klass.__ownedType__.lower() + 's'], event=event)

    klass.__setstate__ = setstate

    def getcopystate(instance):
        try:
            state = super(klass, instance).__getcopystate__()
        except AttributeError:
            state = dict()
        state['%ss' % klass.__ownedType__.lower()] = \
            [ownedObject.copy() for ownedObject in objects(instance)]
        return state

    klass.__getcopystate__ = getcopystate

    return klass



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''


class ColorMixer(object):
    @staticmethod    
    def mix(colors):
        colorSums, colorCount = [0, 0, 0, 0], 0
        for color in colors:
            if color:
                try:
                    color = color.Get(includeAlpha=True)
                except AttributeError:
                    pass # color is already a tuple
                for colorIndex in range(len(color)): 
                    colorSums[colorIndex] += color[colorIndex]
                colorCount += 1
        return tuple(colorSum/colorCount for colorSum in colorSums) if colorCount else None
        
